

__all__ = ['Imaging2', 'args']


import numpy as np
from matplotlib import pyplot as plt
import h5py

from multiprocessing import Process, Queue

from .image import Imaging, _get_start_stop

try:
    import bottleneck as bn
    MEDIAN = bn.median
    NANMEDIAN = bn.nanmedian
    MEAN = bn.nanmean
    NANMEAN = bn.nanmean
    SUM = bn.nansum
    NANSUM = bn.nansum
    STD = bn.nanstd
except ImportError:
    MEDIAN = np.median
    NANMEDIAN = np.nanmedian
    MEAN = np.mean
    NANMEAN = np.nanmean
    SUM = np.sum
    NANSUM = np.nansum
    STD = np.std


class Imaging2(Imaging):

    def fill_grid(self):
        super().fill_grid()
        # sort by self.ind_g then by self.ind_cata
        inds = np.lexsort((self.ind_cata, self.ind_g))
        self.ind_cata = self.ind_cata[inds]
        self.ind_g = self.ind_g[inds]
        self.d2d = self.d2d[inds]
        # d3d if needed

    def set_r_cut(self,):
        args = self.args
        if args.method in ['gaussian', 'bessel_gaussian', 'sinc_gaussian']:
            super().set_r_cut()
        elif args.method in ['wmean', 'wmedian']:
            # use gaussian weights
            method_bak = args.method
            args.method = 'gaussian'
            super().set_r_cut()
            args.method = method_bak
        elif args.method in ['mean', 'median']:
            # no weights needed
            if args.r_cut is None:
                raise(ValueError(f'The method `{args.method}` requires the user to input r_cut'))
            else:
                print(f'r_cut: {args.r_cut} arcsec')

    def _process_sfs(self,):
        """
        add extra info to specfiles for speeding _load_spec_f1
        """
        for sf in self.specfiles:
            if sf.ind_use_chan is not None:
                if (np.diff(sf.ind_use_chan) == 1).all():
                    sf.chan_use_sli = slice(sf.ind_use_chan[0],sf.ind_use_chan[-1]+1)
                else:
                    raise(ValueError('freq shoud be sliced in continuous way'))
            if sf.is_use_t is not None:
                sf.ind_use_t = np.where(sf.is_use_t)[0]

    @staticmethod
    def _load_spec_f1(sf, inds, key, polar, scale_beams=None):
        """
        sf
        inds: indices after limit ra and dec range
        """
        f = h5py.File(sf.fpath)
        f_spec = f['S'][key]
        if polar == 'XX':
            sli = 0
        elif polar == 'YY':
            if f_spec.shape[0] < 2:
                raise(ValueError(f'polar method is YY, but spectra have only one polar.'))
            sli = 1
        else:
            sli = slice(None)
        if sf.is_use_t is not None:
            ind_use_ori = sf.ind_use_t[inds]
        else:
            ind_use_ori = inds

        if sf.ind_use_chan is not None:
            _specs = f_spec[sli, ind_use_ori, sf.chan_use_sli]
        else:
            _specs = f_spec[sli, ind_use_ori]

        if polar == 'M':
            _specs = np.mean(_specs, axis=0) # only two float32 to average
        elif polar in ['XX', 'YY']:
            ...
        elif polar == 'B':
            ...
        else:
            raise(ValueError(f'polar method not support {polar}'))
        if scale_beams is not None:
            _specs *= scale_beams[sf.nB-1]
            print(self.scale_beams[sf.nB-1])
        return _specs



    def _load_spec(self, s, e):
        """
        load spec for one grid from `self.specfiles`
        """
        args = self.args
        ifile_ = self.ifile[s:e]
        ind_ifile_ = self.ind_ifile[s:e]
        # ifile_ and ind_ifile_ is sorted in ifile_ order and then in ind_ifile_ order after
        # self.ind_g and self.ind_cata (and ...) sorted by self.ind_g then by self.ind_cata
        ifile_uni, ifile_num = np.unique(ifile_, return_counts=True)
        ind_ifile_list = np.split(ind_ifile_, np.cumsum(ifile_num)[:-1])
        specs = np.vstack([self._load_spec_f1(self.specfiles[i], ii, args.key, args.polar)
                          for i, ii in zip(ifile_uni,ind_ifile_list)])

        return specs

    @staticmethod
    def calc(q, obj, start_, stop_, ind_g_uni_, verbose=False):
        args = obj.args

        _MEAN = NANMEAN
        _MEDIAN = NANMEDIAN
        _SUM = NANSUM

        old_set = np.seterr()
        np.seterr(invalid='ignore')

        iter_ = zip(start_, stop_, ind_g_uni_)
        if verbose:
            from tqdm import tqdm
            iter_ = tqdm(iter_, total=len(start_), desc='CPU 0: ', mininterval=2)

        for s, e, ii in iter_:
            specs = obj._load_spec(s, e)
            # deal with nan value
            is_finite = np.isfinite(specs)
            num = len(specs)
            num_finite_chan = np.sum(is_finite, axis=0)

            if args.method in ['gaussian', 'bessel_gaussian', 'sinc_gaussian']:
                # convolve
                weis = obj.conv_weis[s:e]
                weis = weis.reshape((-1,)+(1,)*(specs.ndim-1))
                weis_sum_chan = _SUM(weis*is_finite, axis=0)
                spec_g = _SUM(specs.astype('float64')*weis, axis=0) / weis_sum_chan
                # `weis_sum_chan` only those methods
                obj.pixel_weis_sum_chan[ii] = weis_sum_chan
            elif args.method in ['wmean', 'wmedian']:
                weis = obj.conv_weis[s:e]
                weis = weis.reshape((-1,)+(1,)*(specs.ndim-1))
                if args.method == 'wmean':
                    STAT = _MEAN
                else:
                    STAT = _MEDIAN
                spec_g = STAT(specs.astype('float64')*weis, axis=0) / STAT(weis*is_finite, axis=0)
            elif args.method == 'mean':
                spec_g = _MEAN(specs.astype('float64'), axis=0)
            elif args.method == 'median':
                spec_g = _MEDIAN(specs.astype('float64'), axis=0)
            # limit nan frac
            spec_g[num_finite_chan < num*args.frac_finite_min] = np.nan
            # change spec inplace
            obj.pixel_data[ii] = spec_g
            obj.pixel_specs_nums[ii] = num
            obj.pixel_finite_nums_chan[ii] = num_finite_chan

        np.seterr(**old_set)
        if verbose:
            print('Waiting other CPUs')
#         res =
#         if q is None:
#             # not multiprocessing
#             return res
#         else:
#             q.put(res)


    def __call__(self):
        args = self.args
        self.gen_fpaths()
        self.gen_scale_beams() # also change args.beam_fwhw
        self.set_r_cut()
        self.gen_specfiles()
        self.check_spec_key()
        self.check_axis3()
        self.gen_header_grid()
        self.fill_grid()


        if args.method in ['gaussian', 'bessel_gaussian', 'sinc_gaussian']:
            # convolve
            self.gen_conv_weis()
        elif args.method in ['wmean', 'wmedian']:
            # use gaussian weights
            method_bak = args.method
            args.method = 'gaussian'
            self.gen_conv_weis()
            args.method = method_bak
        elif args.method in ['mean', 'median']:
            # no weights needed
            pass
        else:
            raise(ValueError(f'method {args.method} not support'))

        self._process_sfs()
        # output array only stored the results and don't has other operation, use 'float32'
        self.init_out(DataType='float32')
        # data init as nan not zeros
        self.pixel_data[:] = np.nan
        self.pixel_weis_sum_chan[:] = np.nan

        # determine the file and indices of the spec in
        #
        nspecs = np.array([len(sf.ra) for sf in self.specfiles])
        lens_cum = np.hstack([0, np.cumsum(nspecs)])
        self.ifile =  np.searchsorted(lens_cum, self.ind_cata + 1, side='left') - 1
        self.ind_ifile = self.ind_cata - lens_cum[self.ifile]
        # self.ifile, self.ind_ifile, self.ind_g, self.inds_cata, self.d2d
        #
        # find the start and stop index for each grid
        ind_g_uni = np.unique(self.ind_g)
        start, stop = _get_start_stop(ind_g_uni, self.ind_g)
        # distribute to each process
        ind_g_uni_list = np.array_split(ind_g_uni, args.nproc)
        start_list = np.array_split(start, args.nproc)
        stop_list = np.array_split(stop, args.nproc)
        ps = []
        qs = []
        verbose = True
        for start_, stop_, ind_g_uni_ in zip(start_list, stop_list, ind_g_uni_list):
                q = Queue()
                p = Process(target=self.calc, args=(q, self, start_, stop_, ind_g_uni_, verbose))
                ps += [p]
                qs += [q]
                verbose = False
        for p in ps:
            p.start()
        #self.res = [q.get() for q in qs]
        for p in ps:
            p.join()  # need after q.get()
        #print('gathering results')

        # saving

        _shape = self.ra_grid.shape + self.pixel_data.shape[-1:]
        self.data_cube = self.pixel_data.reshape(_shape).transpose(2,0,1)
        self.weis_chan_cube = self.pixel_weis_sum_chan.reshape(_shape).transpose(2,0,1)
        self.nums_chan_cube = self.pixel_finite_nums_chan.reshape(_shape).transpose(2,0,1)
        self.save()


from argparse import Namespace

## exampler parameters
args = Namespace()

args.ra_range = [20.254, 27.25]
args.dec_range = None

args.wcs_ra_center = None
args.wcs_dec_center = None

args.range3 = [-2000, 1500]
args.type3 = 'vrad'
args.bwidth = [60,]
args.proj = 'SIN'
args.r_cut = 90 # arcsec
##
args.method = 'bessel_gaussian'
args.beam_fwhw = 2.9 # arcmin; beam Full width at half maximum
args.gaussian_fwhw = None
args.bsize = None # for Bessel in Bessel*gauss or sin in sin*gauss
args.gsize = None # for gauss in Bessel*gauss or sin*gauss

args.fpatterns = ['/home/jyj/jingyj/pipeline_test/M31_v4_highz/S2/med/*-bld-fc-ds.hdf5', ]

args.outname = 'test.fits'
args.force = True

args.nproc = 5
args.step = 19
args.share_mem = True

args.key = None
args.wcs_from = '/home/jyj/jingyj/M33/M33_local_areciob.fits'

args.scale_beams_file = None

args.polar = 'M'

args.frac_finite_min = 0.01



#!/usr/bin/env python
# coding: utf-8

import numpy as np
from astropy import log
from hifast.utils.misc import down_sample
import os
import sys

def down(properties,keys,ra,dec,mjd,freq,vel,chanfactor=None,specfactor=None):
    
    if specfactor is not None:
        dec=down_sample(dec[None,:,None],specfactor)[0,:,0]
        ra=down_sample(ra[None,:,None],specfactor)[0,:,0]
        mjd=down_sample(mjd[None,:,None],specfactor)[0,:,0]
        for key in keys:
            properties[key]=down_sample(properties[key].T[:,:,None],specfactor)[:,:,0].T
        print(f"After spec down={properties[key].shape}")
    
    if chanfactor is not None:
        vel= down_sample(vel[None,:,None],chanfactor)[0,:,0]
        freq=down_sample(freq[None,:,None],chanfactor)[0,:,0]
        for key in keys:
            properties[key]=down_sample(properties[key][:,:,None],chanfactor)[:,:,0]
        print(f"After channel down={properties[key].shape}")
    return properties,ra,dec,mjd,freq,vel
    
if __name__ == '__main__':
    
    import argparse
    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('fname',
                        help='file name')
    parser.add_argument('--outdir',
                       help='default is same with the input file')

    parser.add_argument('--key', nargs='+',
                       help='properties to down sample, flux, Ta or both')

    parser.add_argument('--chanfactor',type=int,
                       help=' down sample on channel')

    parser.add_argument('--specfactor',type=int,
                       help=' down sample on spec')

    parser.add_argument('-f', '--force', action='store_true',
                       help='overwriting file if out file exists')

    args = parser.parse_args()
    fname= args.fname
    outdir= args.outdir
    keys= args.key
    chanfactor= args.chanfactor
    specfactor= args.specfactor

    #save
    if outdir is None:
        outdir= os.path.dirname(fname)
    fileout= os.path.join(outdir, '.'.join(os.path.basename(fname).split('.')[:-1]) +'-ds.hdf5' )

    if os.path.exists(fileout):
        if args.force:
            print(f"will overwrite the existing out file {fileout}")
        else:
            print(f"File exists {fileout}")
            print('exit... Using -f to overwrite it.')
            sys.exit()

    import h5py
    f= h5py.File(fname,'r')
    properties={}
    if keys is None:
        keys=[]
        for key in ['Ta', 'flux', ]:
            if key in f.keys():
                keys+=[key]
    for key in keys:
        T = f[key][()]
        if len(T.shape) == 3:
            T = np.mean(T, axis=2, dtype='float64') 
        properties[key]= T
    print(f"Origin shape={properties[key].shape}")

    freq= f['freq'][()]
    mjd= f['mjd'][()]
    ra= f['ra'][()]
    dec= f['dec'][()]
    vel= f['vel'][()]

    if 'Header' in f.keys():
        from collections import OrderedDict
        header_in= OrderedDict(f['Header'].attrs.items())
    else:
        header_in=None


    #down sample
    properties,ra,dec,mjd,freq,vel = down(properties,keys,ra,dec,mjd,freq,vel,chanfactor,specfactor)

    dict_out={}
    dict_out['freq']=freq
    dict_out['vel']=vel
    dict_out['ra']=ra
    dict_out['dec']=dec
    dict_out['mjd']=mjd
    for key in properties.keys():
        dict_out[key]= properties[key].astype('float32')

    from hifast.utils.io import  rec_his, save_dict_hdf5
    f.close()
    header=rec_his(args=args)
    
    if header_in is not None: header.update(header_in)
    save_dict_hdf5(fileout, dict_out, header=header)
    log.info(f"Saved to {fileout}")





__all__ = ['sep_line', 'parser', 'IO']



from .utils.io import *

sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='T Cal, Src - Ref', )
parser.add_argument('fpath',
                    help='only need input the first chunk file path (e.g. XXX_0001.fits) of RAW spectra data')
parser.add_argument('-f', dest='force', action='store_true',
                    help='overwriting file if out file exists')
parser.add_argument('--outdir', required=True,
                   help='the directory to store output files.')
parser.add_argument('-g', is_write_out_config_file_arg=True,
                        help='save config to file path')
parser.add_argument('-c', '--my-config', is_config_file_arg=True,
                        help='config file path')
parser.add_argument('--frange', type=float, nargs=2,
                    help='freq range')


parser.add_argument('-d', '--n_delay', type=int, required=True,
                   help='time of delay divided by sampling time')
parser.add_argument('-m', '--n_on', type=int, required=True,
                   help='time of Tcal_on divided by sampling time')
parser.add_argument('-n', '--n_off', type=int, required=True,
                    help='time of Tcal_off divided by sampling time')

parser.add_argument('--smooth', choices=['mean','poly','gaussian'], required=True,
                    help="smooth method: 'mean','poly',gaussian',.. default:gaussian")
parser.add_argument('--s_sigma', type=float, default=5,
                    help='sigma for gaussian smooth, default 5MHz')

parser.add_argument('--noise_mode', default='high', choices=['high','low'],
                    help='noise_mode, high or low')
parser.add_argument('--noise_date', default='auto',
                    help='noise obs date, default auto')
parser.add_argument('--med_filter_size_cal', type=int, default=5,
                    help='median filter kernel size for power of cal; odd number; default 5')


parser.add_argument('--t_src', type=float, required=True,
                   help='on-source time[second]')
parser.add_argument('--t_ref', type=float, required=True,
                   help='off-source time[second]')
parser.add_argument('--t_change', type=float, required=True, choices=[30,60],
                    help="switching time[second]. 30s for sepatation of src and ref less than 20'; 60s for sepatation between 20' and 60'.")
parser.add_argument('--n_repeat', type=int, required=True,
                    help='The number of on-source off-source cycles')

parser.add_argument('--only_off', type=bool_fun, choices=[True, False], default='False',
                    help='only noise off')



class IO(Path_IO):
    def __init__(self, args, inplace_args=False):
        """
        args: class
              including attributes: fpath, outdir, frange
        """
        self.args = args if inplace_args else copy.deepcopy(args)
        self._gen_fpath_out()
        self._check_fout()
        self._import_m()
        self.nB = self.get_nB(self.args.fpath)
        self.load_and_add_Header()

    def _import_m(self,):
        global np
        import numpy as np

    def _gen_fpath_out(self,):
        """
        """
        args = self.args

        # replace patten in outdir
        nB = self.get_nB(args.fpath)
        project = get_project(args.fpath)
        ## use the dirname of the fits file as "date"
        date = os.path.basename(os.path.dirname(os.path.abspath(args.fpath)))
        args.outdir = sub_patten(args.outdir, date=date, nB=f'{nB:02d}', project=project)
        args.outdir = os.path.expanduser(args.outdir)
        print(f'outdir: {args.outdir}')
        if not os.path.exists(args.outdir):
            print(f'outdir {args.outdir} not exists. Create it now')
            os.makedirs(args.outdir, exist_ok=True)

        fname_part = re.sub('[0-9]{4}\.fits\Z', '', args.fpath)
        fname_add = os.path.basename(os.path.dirname(os.path.abspath(fname_part)))
        out_name_base = os.path.join(args.outdir, f"{os.path.basename(fname_part)[:-1]}-{fname_add}")
        self.out_name_base = out_name_base
        fpart = '-S_ps'
        self.fpath_out = f'{out_name_base}{fpart}.hdf5'

    def load_and_add_Header(self, ):
        import json
        Header = {}
        Header.update(rec_his(args=json.dumps(self.args.__dict__)))
        self.Header = Header

    def gen_S(self,):

        from .core.cal import PositionSwitch

        args = self.args
        S = PositionSwitch(args.fpath, args.n_delay, args.n_on, args.n_off,
                 frange=args.frange, verbose=True,
                 smooth=args.smooth, s_para={'s_sigma':args.s_sigma},
                 noise_mode=args.noise_mode, noise_date=args.noise_date,
                 med_filter_size_cal=args.med_filter_size_cal)

        S.sep(args.t_src, args.t_ref, args.n_repeat, args.t_change)
        # S.gen_out_name_base(args.outdir)

        self.S = S

    def __call__(self,):
        args = self.args
        self.gen_S()
        S = self.S

        S.plot = True
        S.out_name_base = self.out_name_base
        S.plot_sep()

        S.gen_Ta(only_off=args.only_off)
        S.gen_radec()
        S.plot_radec(outname=self.fpath_out + '-radec.png')

        dict_out = {}
        dict_out['Header'] = self.Header
        dict_out['Ta'] = MjdChanPolar_to_PolarMjdChan(S.Ta)
        dict_out['freq'] = S.freq_use

        dict_out['ra'] = S.ra
        dict_out['dec'] = S.dec
        dict_out['mjd'] = S.mjd

        # for compatible

        dict_out['is_delay'] = np.array([False,])
        dict_out['is_on'] = np.array([False,])
        dict_out['next_to_cal'] = np.array([False,])
        dict_out['Tcal'] = S.Tcal_s


        self.outfield = 'Ta'
        self.dict_out = dict_out

        print("Saving...")
        save_specs_hdf5(self.fpath_out, self.dict_out, wcs_data_name=self.outfield)
        print(f"Saved to {self.fpath_out}")



if __name__ == '__main__':
    args_ = parser.parse_args()

    # print(parser.format_help())
    # print("----------")
    # print(parser.format_values())  # useful for logging where different settings came from
    print('#'*35+'Args'+'#'*35)
    print(parser.format_values())  # useful for logging where different settings came from
    print('#'*35+'####'+'#'*35)
    io = IO(args_)
    io()




"""
HiFAST Global Configuration Module
----------------------------------

This module handles global configuration management for the HiFAST package.
It supports layered configuration with the following priority (highest first):

1. **Environment Variables**: `HIFAST_<SECTION>_<OPTION>` (e.g., `HIFAST_TCAL_SOURCE`)
2. **User Config File**: `~/.config/hifast/config.ini` (standard INI format)
3. **Defaults**: Hardcoded defaults within this module.

Usage for Developers
--------------------
To add configuration for a new module (e.g., 'fitting'):

1. **Register the Option**:
   Add your new option to `_KNOWN_OPTIONS` dictionary below. This serves as
   direct documentation and enables the `list` CLI command.
   
   _KNOWN_OPTIONS = {
       ...
       'fitting': {
           'max_iter': {'default': 50, 'help': "Maximum iterations for fitter"},
           'algorithm': {'default': 'gaussian', 'help': "Fitting algorithm to use"}
       }
   }

2. **Access in Code**:
   Import `conf` and use typed getters:
   
   from hifast.utils.config import conf
   max_iter = conf.get_int('fitting', 'max_iter', 50)
   algo = conf.get('fitting', 'algorithm', 'gaussian')

"""

import os
import sys
import configparser
from pathlib import Path

# Constants for Defaults
_TCAL_GITHUB_MANIFEST = "https://raw.githubusercontent.com/jyingjie/hifast-tcal-data/main/manifest.json"
_TCAL_GITHUB_BASE = "https://github.com/jyingjie/hifast-tcal-data/releases/download"

_TCAL_R2_MANIFEST = "https://pub-b2347aee569d45789e90c47ad3177e0e.r2.dev/tcal-data/manifest.json"
_TCAL_R2_BASE = "https://pub-b2347aee569d45789e90c47ad3177e0e.r2.dev/tcal-data"

# Known Options Registry
# ----------------------
# DOCUMENTATION SOURCE OF TRUTH.
# Add new module parameters here to expose them to users via `hifast.utils.config list`.
_KNOWN_OPTIONS = {
    'general': {
        'offline': {
            'default': False,
            'help': "Enable offline mode. Disable all network requests. (Env: HIFAST_GENERAL_OFFLINE)"
        }
    },
    'tcal': {
        'source': {
            'default': 'r2', 
            'help': "Data source used for Tcal files. Options: 'r2' (default, Cloudflare) or 'github'."
        },
        'tcal_dir': {
            'default': '~/Tcal/', 
            'help': "Local directory where Tcal data is stored/cached."
        },
        'manifest_url': {
            'default': None, 
            'help': "Remote Manifest URL. Required if source='custom'. Can also be used to override 'r2'/'github' defaults."
        },
        'base_url': {
            'default': None, 
            'help': "Remote Base URL. Required if source='custom'. Can also be used to override 'r2'/'github' defaults."
        },
    }
}

_TCAL_SOURCES = {
    'github': {'manifest': _TCAL_GITHUB_MANIFEST, 'base': _TCAL_GITHUB_BASE},
    'r2': {'manifest': _TCAL_R2_MANIFEST, 'base': _TCAL_R2_BASE}
}

# XDG Config Home or ~/.config
CONFIG_DIR = Path(os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")) / "hifast"
CONFIG_FILE = CONFIG_DIR / "config.ini"

class ConfigManager:
    def __init__(self):
        self._config = configparser.ConfigParser()
        self._load()

    def _load(self):
        """Loads configuration from file."""
        if CONFIG_FILE.exists():
            try:
                self._config.read(CONFIG_FILE)
            except Exception as e:
                print(f"Warning: Failed to load config file {CONFIG_FILE}: {e}", file=sys.stderr)
        
        # Ensure default sections exist
        if 'tcal' not in self._config:
            self._config['tcal'] = {}
        if 'general' not in self._config:
            self._config['general'] = {}

    def save(self):
        """Saves current configuration to file."""
        try:
            CONFIG_DIR.mkdir(parents=True, exist_ok=True)
            with open(CONFIG_FILE, 'w') as f:
                self._config.write(f)
        except Exception as e:
            print(f"Error saving config file: {e}", file=sys.stderr)

    def get(self, section, option, default=None):
        """
        Get value with priority:
        1. Environment Variable (HIFAST_SECTION_OPTION)
        2. Config File
        3. Default
        """

        env_var = f"HIFAST_{section.upper()}_{option.upper()}"
        if env_var in os.environ:
            return os.environ[env_var]
        
        return self._config.get(section, option, fallback=default)

    def get_int(self, section, option, default=None):
        """Get integer value with environment override."""
        val = self.get(section, option, default)
        try:
            return int(val)
        except (ValueError, TypeError):
            return default

    def get_float(self, section, option, default=None):
        """Get float value with environment override."""
        val = self.get(section, option, default)
        try:
            return float(val)
        except (ValueError, TypeError):
            return default

    def get_boolean(self, section, option, default=None):
        """Get boolean value with environment override."""
        val = self.get(section, option, default)
        if val is None:
            return default
        if isinstance(val, bool):
            return val
        s = str(val).lower()
        if s in ('true', 'yes', 'on', '1'):
            return True
        if s in ('false', 'no', 'off', '0'):
            return False
        return default

    def set(self, section, option, value, save=False):
        """Sets a value in memory, optionally saving to file."""
        if section not in self._config:
            self._config[section] = {}
        self._config[section][option] = str(value)
        if save:
            self.save()

    def get_tcal_urls(self):
        """
        Resolves Tcal URLs based on configuration.
        Returns: (manifest_url, base_url)
        """
        # 1. Determine base defaults from source
        source = self.get('tcal', 'source', 'r2')
        
        preset = _TCAL_SOURCES.get(source, {})
        default_manifest = preset.get('manifest')
        default_base = preset.get('base')
        
        if not default_manifest and source != 'custom':
             print(f"Warning: Unknown [tcal] source='{source}'. Using GitHub default.", file=sys.stderr)
             default_manifest = _TCAL_GITHUB_MANIFEST
             default_base = _TCAL_GITHUB_BASE

        # 2. Check for explicit overrides (Env or Config)
        # These take precedence over source-derived defaults
        manifest_url = self.get('tcal', 'manifest_url')
        base_url = self.get('tcal', 'base_url')

        return (manifest_url or default_manifest), (base_url or default_base)

# Global instance
# Global instance
conf = ConfigManager()

def get_parser():
    import argparse
    import textwrap
    
    description = textwrap.dedent("""
    Manage HiFAST Global Configuration
    ==================================
    
    Set and retrieve configuration values for HiFAST modules (e.g., Tcal).
    
    Priority Order:
      1. Environment Variables:
         - Standard: HIFAST_SECTION_OPTION (e.g., HIFAST_TCAL_SOURCE)
      2. Config File (~/.config/hifast/config.ini)
      3. Global Defaults
    """)
    
    epilog = textwrap.dedent("""
    Examples:
      # List all options
      python -m hifast.utils.config list
      
      # Set Tcal source to R2 (default)
      python -m hifast.utils.config set tcal source r2
      
      # Get current data directory
      python -m hifast.utils.config get tcal data_dir
    """)

    parser = argparse.ArgumentParser(
        description=description, 
        epilog=epilog, 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    subparsers = parser.add_subparsers(dest='command')
    
    # set command
    set_parser = subparsers.add_parser('set', help='Set a configuration value')
    set_parser.add_argument('section', help='Config section (e.g., tcal)')
    set_parser.add_argument('option', help='Config option (e.g., source)')
    set_parser.add_argument('value', help='Value to set')
    
    # get command
    get_parser_cmd = subparsers.add_parser('get', help='Get a configuration value')
    get_parser_cmd.add_argument('section', help='Config section')
    get_parser_cmd.add_argument('option', help='Config option')

    # list command
    subparsers.add_parser('list', help='List all available options and current values')
    
    return parser

def main():
    parser = get_parser()
    args = parser.parse_args()
    
    if args.command == 'set':
        # Validation checks
        if args.section in _KNOWN_OPTIONS:
            valid_opts = _KNOWN_OPTIONS[args.section].keys()
            if args.option not in valid_opts:
                import sys
                import difflib
                print(f"Warning: '{args.option}' is not a known option for section '{args.section}'.", file=sys.stderr)
                matches = difflib.get_close_matches(args.option, valid_opts)
                if matches:
                    print(f"Did you mean: {', '.join(matches)}?", file=sys.stderr)
                print(f"Valid options: {', '.join(valid_opts)}", file=sys.stderr)
        else:
             import sys
             print(f"Warning: '{args.section}' is not a known section.", file=sys.stderr)
             print(f"Known sections: {', '.join(_KNOWN_OPTIONS.keys())}", file=sys.stderr)

        conf.set(args.section, args.option, args.value, save=True)
        print(f"Set [{args.section}] {args.option} = {args.value}")
    elif args.command == 'get':
        # Lookup default from KNOWN_OPTIONS if available
        default = None
        if args.section in _KNOWN_OPTIONS and args.option in _KNOWN_OPTIONS[args.section]:
            default = _KNOWN_OPTIONS[args.section][args.option].get('default')
            
        val = conf.get(args.section, args.option, default)
        print(val)
    elif args.command == 'list':
        import textwrap
        col_sec = 10
        col_opt = 15
        col_env = 30
        col_val = 30
        col_desc = 35 # Allow simple wrap
        
        # Header
        print(f"{'SECTION':<{col_sec}} {'OPTION':<{col_opt}} {'ENV VAR':<{col_env}} {'VALUE':<{col_val}} {'DESCRIPTION'}")
        print("-" * 130)
        
        # 1. Print Known Options
        for section, options in _KNOWN_OPTIONS.items():
            for opt, meta in options.items():
                val = str(conf.get(section, opt, meta.get('default')))
                desc = meta.get('help', '')
                env_var = f"HIFAST_{section.upper()}_{opt.upper()}"
                
                # Truncate value if too long (rare config)
                if len(val) > col_val - 2:
                    val = val[:col_val-5] + "..."
                
                # Wrap description
                wrapped_desc = textwrap.wrap(desc, width=50) # Use loose width for desc
                
                # First line
                d_first = wrapped_desc[0] if wrapped_desc else ""
                print(f"{section:<{col_sec}} {opt:<{col_opt}} {env_var:<{col_env}} {val:<{col_val}} {d_first}")
                
                # Subsequent lines
                for line in wrapped_desc[1:]:
                     print(f"{'':<{col_sec}} {'':<{col_opt}} {'':<{col_env}} {'':<{col_val}} {line}")
    else:
        parser.print_help()

if __name__ == "__main__":
    main()





__all__ = ['plot_im', 'plot']


import matplotlib.pyplot as plt
import numpy as np
import scipy.interpolate as interp
import h5py


def _tight_ra(ra):
    ra_s= np.sort(ra)
    diff_s= np.diff(ra_s)
    ind_max= np.argmax(diff_s)
    if diff_s[ind_max] < ra_s[0]+360-ra_s[-1]:
        return ra
    else:
        ra= np.copy(ra)
        is_c= ra>=ra_s[ind_max+1]
        ra[is_c]= ra[is_c]-360
        return ra


def plot_im(vals, x=None, y=None, x_plot_range=None, y_plot_range=None, ax=None, vlines=None, colorbar=True, y2=None,
         vmin='per0.01', vmax='per95', **kwargs):
    """
    vals: shape (m,n)
    x: None or shape (n,)
    y: None or shape (m,)
    """

    if x is None:
        x = np.arange(vals.shape[1])
    if y is None:
        y = np.arange(vals.shape[0])

    if x_plot_range is not None:
        is_ = (x >= x_plot_range[0]) & (x <= x_plot_range[1])
        x = x[is_]
        vals = vals[:,is_]
    if y_plot_range is not None:
        is_ = (y >= y_plot_range[0]) & (y <= y_plot_range[1])
        y = y[is_]
        vals = vals[is_,:]
        if y2 is not None:
            y2 = y2[is_]
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(10,10))
#     if imshow_kwargs == {}:
#         imshow_kwargs = {}
    else:
        fig = None
    # convert to value
    if vmax is not None:
        vmax = float(vmax) if 'per' != str(vmax)[:3] else np.nanpercentile(vals, float(str(vmax)[3:]), interpolation='nearest')
    if vmin is not None:
        vmin = float(vmin) if 'per' != str(vmin)[:3] else np.nanpercentile(vals, float(str(vmin)[3:]), interpolation='nearest')

    im = ax.imshow(vals[:], aspect='auto', extent=(x[0], x[-1], y[0], y[-1]), origin='lower', vmin=vmin, vmax=vmax, **kwargs)
    #rasterized=False
    if vlines is not None:
        for vline in vlines:
            ax.axvline(x=vline, color='r', linestyle=':')
    if y2 is not None:
        try:
            def forward(x):
                return interp.interp1d(y, y2, fill_value='extrapolate')(x)
            def inverse(x):
                return interp.interp1d(y2, y, fill_value='extrapolate')(x)
            ax.secondary_yaxis('right', functions=(forward, inverse))
        except:
            pass
    ax.minorticks_on()
    if colorbar:
        plt.colorbar(im, ax=ax)
    return im, fig


def plot(fname, ax, polar=0, ytick1='index', ytick2=None, replace_rfi = False, **kwargs):
    """
    fname:
    ax:
    polar: polarization, 0 or 1
    ytick1: str; index, ra, dec, mjd or time
    ytick2: str; index, ra, dec, mjd or None
    """
    f = h5py.File(fname,'r')
    S = f['S']
    if 'T' in S.keys():
        vals = S['T']
    elif 'Ta' in S.keys():
        vals = S['Ta']
    elif 'flux' in S.keys():
        vals = S['flux']
    else:
        raise(ValueError('can not find spec'))

    if 'vel' in S.keys():
        x = S['vel'][()]
    elif 'freq' in S.keys():
        x = S['freq'][()]
    else:
        raise()

    if vals.ndim == 3:
        if polar == -1:
            vals = np.mean(vals, axis = 0)
        else:
            vals = vals[polar]

    if replace_rfi:
        if 'is_rfi' in S.keys():
            is_rfi = S['is_rfi'][:]
            vals[is_rfi] = np.nan

    if ytick1 in ['ra', 'dec', 'mjd', 'time']:
        key = 'mjd' if ytick1=='time' else ytick1
        try:
            y = S[key][()]
        except Exception as err:
            print(err)
    elif ytick1 == 'index':
        y = np.arange(vals.shape[1])
    y2 = None
    if ytick2 in ['ra', 'dec', 'mjd', 'time']:
        key = 'mjd' if ytick2=='time' else ytick2
        try:
            y2 = S[key][()]
        except Exception as err:
            print(err)
    elif ytick2 == 'index':
        y2 = np.arange(vals.shape[1])

    im, _ = plot_im(vals, x, y=y, y2=y2, ax=ax, **kwargs)
    if ytick1=='time':
        from astropy.time import Time, TimezoneInfo
        import astropy.units as u
        labels_new = ((Time(ax.axes.get_yticks(), format='mjd')) + 8*u.hour).strftime('%H:%M:%S')
        ax.axes.set_yticklabels(labels_new)
    return im

def _check_fout(args, fpath_out):
    """
    check whether fout exists
    """
    ret = None
    if os.path.exists(fpath_out):
        if args.force:
            print(f"will overwrite the existing output file {fpath_out}")
        else:
            print(f"File exists {fpath_out}")
            print("exit... Use ' -f ' to overwrite it.")
            ret = "exit"
    return ret


if __name__ == '__main__':
    import argparse
    import os
    import re

    parser = argparse.ArgumentParser(allow_abbrev=False)
    parser.add_argument('fnames', nargs='+',
                        help='fnames')
    parser.add_argument('--outdir', default='./',
                       help='default is ./')
    parser.add_argument('-s', '--single', action='store_true',
                       help="")
    parser.add_argument('-f', action='store_true', dest='force',
                        help='if set, overwriting file if output file exists')
    parser.add_argument('--xrange', type=float, nargs=2,
                       help='x axis range')
    parser.add_argument('--vmin', type=str, default='per0.1',
                       help='')
    parser.add_argument('--vmax', type=str, default='per95',
                       help='')
    parser.add_argument('--interpolation', type=str, default='nearest',
                       help="'none', 'antialiased', 'nearest', 'bilinear', 'bicubic', 'spline16', 'spline36',")
    parser.add_argument('--ytick1', choices=['index', 'ra', 'dec', 'mjd', 'time'], default='index',
                       help='')
    parser.add_argument('--ytick2', choices=['index', 'ra', 'dec', 'mjd'],
                       help='')
    parser.add_argument('--vlines', type=float, nargs='+',
                       help='')
    parser.add_argument('--show', action='store_true',
                       help='')
    parser.add_argument('--replace_rfi', action='store_true',
                       help='rfi')
    parser.add_argument('--polar', type=int, default=0,
                       help='polar')

    args = parser.parse_args()
    fnames = args.fnames
    vlines = args.vlines
    outdir = args.outdir
    xrange = args.xrange

    if not args.show:
        plt.switch_backend('agg')
    fnames.sort()
    single = args.single
    if len(fnames)==1:
        single = True
    rasterized = True if args.interpolation == 'none' else False
    imshow_kwargs = {}
    imshow_kwargs['interpolation'] = args.interpolation

    from tqdm import tqdm
    if not single:
        from itertools import groupby
        
        #keys = list(map(lambda x: re.sub('-M[0-1][0-9]','-M00', os.path.basename(x).split('-specs_T')[0]), fnames))
        keys = list(map(lambda x: re.sub('-M[0-1][0-9]','-M00', os.path.basename(x)), fnames))
        
        # Sort by keys for groupby
        combined = sorted(zip(keys, fnames), key=lambda x: x[0])
        
        for key, group in tqdm(groupby(combined, lambda x: x[0])):
            # group is iterator of (key, fname)
            _files_fname = [x[1] for x in group]
            fpath_out = f'{outdir}/{key}.19.pdf'
            ret = _check_fout(args, fpath_out)
            if ret == "exit": continue

            nrows=4
            ncols=5
            fig, axs = plt.subplots(nrows, ncols, figsize=(160/3,90/3), sharex=True, sharey=True)
            axs = axs.flatten()
            for i, (fname, ax) in enumerate(zip(_files_fname, axs[:19])):
                im = plot(fname, ax, polar=args.polar, ytick1=args.ytick1, ytick2=args.ytick2, vmin=args.vmin, vmax=args.vmax,
                          x_plot_range=xrange, vlines=vlines, colorbar=False,replace_rfi = args.replace_rfi,
                          **imshow_kwargs)
            fig.colorbar(im, ax= axs[-1])
            ax = axs[-1]
            ax.plot([], [], label=key)
            ax.legend()

            fig.tight_layout()
            if args.show:
                fig.show()
                input()
            fig.savefig(fpath_out)
            fig.clear()
    else:
        for fname in tqdm(fnames):
            fbasename = os.path.basename(fname)
            fpath_out = f'{outdir}/{fbasename}.pdf'
            ret = _check_fout(args, fpath_out)
            if ret == "exit": continue

            nrows=1
            ncols=1
            fig, ax = plt.subplots(nrows, ncols, figsize=(15,12), sharex=True, sharey=True)
            im = plot(fname, ax, polar=args.polar, ytick1=args.ytick1, ytick2=args.ytick2, vmin=args.vmin, vmax=args.vmax,
                      x_plot_range=xrange, vlines=vlines, colorbar=False,replace_rfi = args.replace_rfi,
                      **imshow_kwargs)
            fig.colorbar(im, ax= ax)
            ax.set_title(fbasename)

            fig.tight_layout()
            if args.show:
                fig.show()
                input()
            fig.savefig(fpath_out)
            fig.clear()



import os
import warnings

import numpy as np
from scipy import signal
from scipy import ndimage
from scipy.stats import binned_statistic

def apply_along_axis(fun):
    #from functools import wraps
    #@wraps(fun)
    def wrapper(arr, axis, *args, **kwargs):
        """
    arr: array like
    axis: which axis to apply
    ----------------
    Others:"""
        return np.apply_along_axis(fun, axis, arr, *args, **kwargs)
    wrapper.__doc__ += fun.__doc__ 
    wrapper.__name__ = fun.__name__
    wrapper.__qualname__= fun.__qualname__
    return wrapper

@apply_along_axis
def mask_Trues(arr, leng_lim=20):
    """
    mask "continuous Trues" (number >= leng_lim) as True
    Parameters
    ----------
    leng_lim: int
    """
    if not np.any(arr):
        return arr
    arr = np.hstack([[False],arr,[False]])
    diff = np.diff(arr.astype('int32'))
    
    ind_neg = np.where(diff==-1)[0]
    ind_posi = np.where(diff==1)[0]

    leng = (ind_neg - ind_posi)
    is_use = leng < leng_lim 
    
    for i,j in zip(ind_posi[is_use],ind_neg[is_use]):
        arr[i+1:j+1]= False
    return arr[1:-1]

@apply_along_axis
def extend_Trues(is_rfi, leng_lim=1, ext_add=0, ext_frac=1./4):
    """
    extend "continuous Trues" to left and right
    
    Parameters
    ----------
    leng_lim: int; only extend "continuous Trues" which length >= leng_lim
    ext_add: int; fix number to extend
    frac: float; entend length*frac
    """
    if not np.any(is_rfi):
        return is_rfi
    is_rfi = np.hstack([[False],is_rfi,[False]])
    diff= np.diff(is_rfi.astype('int32'))
    
    ind_neg = np.where(diff==-1)[0]
    ind_posi= np.where(diff==1)[0]
#     if ind_neg[0]==0:
#         ind_neg= ind_neg[1:]
#     if ind_posi[-1] == len(diff)-1:
#         ind_posi = ind_posi[:-1]

    leng = (ind_neg- ind_posi)
    is_use= leng>= leng_lim # not needed
    l_ext= np.ceil(leng[is_use]*ext_frac).astype('int') + ext_add
    #l_ext= np.full(len(leng),5)
    for i,n in zip(ind_neg[is_use],l_ext):
        is_rfi[i+1:i+n+1]= True
    for i,n in zip(ind_posi[is_use],l_ext):
        is_rfi[max(0,i-n+1):i+1]= True
    return is_rfi[1:-1]

@apply_along_axis
def median_filter_1d(arr, *args, **kwargs):
    """
    see scipy.ndimage.median_filter
    """
    return ndimage.median_filter(arr, *args, **kwargs)

def median_filter_axis1_d3(arr, kernel_size=5):
    return median_filter_1d(arr, axis=1, size=kernel_size)


def average_every_n(arr, n, axis=-1, drop=True):
    """
    drop: if drop arr.shape[axis]%n in end
    """
    len_ = arr.shape[axis]
    num = len_//n*n
    drop = num==len_ or drop
        
    def _fun(arr_):
        res= np.mean(arr_[:num].reshape((-1,n)), axis=1, dtype='float64')
        if not drop:
            res2 = np.mean(arr_[-n:].reshape((-1,n)), axis=1, dtype='float64')
            res=np.hstack([res,res2])
        return res
    return np.apply_along_axis(_fun, axis=axis, arr=arr)

def down_sample(data,dfactor):
    """
    average every n values along axis 1, n equal dfactor
    Parameters
    ----------
    data : array_like
        shape is (x,N,y)

    axis : int, optional
        The axis in the result array along which the input arrays are stacked.

    dfactor : int
        number of values that take the average

    Returns
    -------
    data_d : ndarray
        shape is (x,N//dfactor,y)

    """
    length=data.shape[1]
    bins=np.arange(0,length+1,dfactor)
    data_d= [binned_statistic(np.arange(length)+0.5,data[:,:,i],'mean',bins=bins)[0] for i in range(data.shape[2])]
    data_d= np.stack(data_d,axis=2)
    return data_d
   

def gaussian_smooth1d(vals, sigma, axis=0):
    '''
    One-dimensional Gaussian smooth.

    Parameters
    ----------
    vals : array_like
        The input array.
    sigma : scalar
        standard deviation for Gaussian kernel
    axis : int, optional
        The axis of `input` along which to calculate. Default is 0.
    '''
    N= min(vals.shape[axis], int(6*sigma))
    #N= vals.shape[axis]
    win_shape= [1 if i!=axis else N  for i in range(len(vals.shape))]
    win_g= signal.windows.gaussian(win_shape[axis],sigma).reshape(win_shape)
    # here,  np.sum(win_g) is same effect with np.sum(win_g,axis=axis) coz other axis shape is 1. 
    res= signal.convolve(vals, win_g, method='fft',mode='same') / np.sum(win_g)      
    return res

def boxcar_smooth1d(vals, sigma, axis=0):
    '''
    One-dimensional moving mean (boxcar) smooth.

    Parameters
    ----------
    vals : array_like
        The input array.
    sigma : scalar
        Number of points = 2*sigma+1
    axis : int, optional
        The axis of `input` along which to calculate. Default is 0.
    '''
    
    win_shape= [1 if i!=axis else 2*int(sigma)+1  for i in range(len(vals.shape))]
    win_g= np.ones(win_shape[axis]).reshape(win_shape)
    #print(win_g)
    res= signal.convolve(vals, win_g, method='fft',mode='same') / np.sum(win_g)
    return res

def smooth_axis1_d3(vals, x=None, method=None, sigma=None, deg=None):
    """
    s_para: dict
    """
    if len(vals) == 0:
        warnings.warn("input array is empty in smooth_axis1_d3")
        return vals
    if vals.ndim !=3:
        raise(ValueError('input array should 3 dim'))
    if method=='poly':
        res= np.zeros_like(vals)
        for i in range(vals.shape[0]):
            for j in range(vals.shape[2]):
                res[i,:,j]= np.poly1d(np.polyfit(x, vals[i,:,j], deg=deg))(x)
        return res
    
    if method=='gaussian':
        extend= min((int(sigma*4), vals.shape[1]))

    #     if vals.shape[1] < extend:
    #         num= int(np.ceil(extend/vals.shape[1]))
    #         tail= np.concatenate([vals,]*1,axis=1)[:,-extend:,:][:,::-1,:]
    #         head= np.concatenate([vals,]*1,axis=1)[:,:extend,:][:,::-1,:]
    #     else:
        tail= vals[:,-extend:,:][:,::-1,:]
        head= vals[:,:extend,:][:,::-1,:]

        vals_new= np.concatenate((head, vals, tail),axis=1)
        res= gaussian_smooth1d(vals_new,sigma,axis=1)
        res= res[:,extend:-extend,:]

        return res

    
def convolve1d(arr, kernel1d, axis=-1, mode='reflect', method='auto'):
    '''
    One-dimensional convolving kernel1d along axis of arr

    Parameters
    ----------
    arr : array_like
        The input array.
    kernel1d : array_like
        dimension = 1
    axis : int, optional
        The axis of `input` along which to calculate. Default is -1.
    mode : str{'reflect', 'none'}, optional
        The mode parameter determines how "arr" is extended beyond its boundaries.
    method : str{'auto', 'direct', 'fft'} 
        see scipy.signal.convolve
    '''
    # boundary
    if mode == 'reflect':
        extend = int(len(kernel1d)/2)
        
        sli_h = [slice(None)]*arr.ndim
        sli_h[axis] = slice(None, extend)
        sli_h = tuple(sli_h)

        sli_t = [slice(None)]*arr.ndim
        sli_t[axis] = slice(-extend,None)
        sli_t = tuple(sli_t)

        sli_flip = [slice(None)]*arr.ndim
        sli_flip[axis] = slice(None, None, -1)
        sli_flip = tuple(sli_flip)

        if extend > arr.shape[axis]:
            n_add_ker = int(np.ceil(extend/arr.shape[axis]))//2*2+1  # odd number
            if n_add_ker < 3:
                n_add_ker = 3
            if n_add_ker > 9:
                raise(ValueError('kernel length is too larger that arr'))
            arr_ = np.concatenate([arr,] + [arr[sli_flip], arr,]*((n_add_ker-1)//2), axis=axis)
        else:
            arr_ = arr    
        tail = arr_[sli_t][sli_flip]
        head = arr_[sli_h][sli_flip]
        arr = np.concatenate((head, arr, tail),axis=axis)
        del(arr_, tail, head)
    elif mode == 'none':
        extend = None
    else:
        raise(ValueError("mode not support"))
    # change kernel shape    
    ker_shape = [1,]*arr.ndim
    ker_shape[axis] = len(kernel1d)
    kernel = kernel1d.reshape(ker_shape)
    # here,  np.sum(kernel) is same effect with np.sum(win_g,axis=axis) coz other axis shape is 1. 
    res = signal.convolve(arr, kernel, method=method, mode='same') / np.sum(kernel)
    if extend is not None:
        sli_res = [slice(None)]*arr.ndim
        sli_res[axis] = slice(extend,-extend)
        sli_res = tuple(sli_res)
        res = res[sli_res]
    return res

def smooth1d(arr, method, sigma, axis=-1):
    '''
    One-dimensional smooth arr along axis using different method

    Parameters
    ----------
    arr : array_like
        The input array.
    method : str{{'gaussian_fft', 'boxcar', 'gaussian', 'median'}}
        'gaussian_fft': use scipy.signal.convolve; sigma is std; extend boundaries using 'reflect'.
        'boxcar': use scipy.signal.convolve; average every int(2*sigma+1); extend boundaries using 'reflect'.
        'gaussian': use scipy.ndimage.gaussian_filter1d; sigma is std
        'median': use scipy.ndimage.median_filter; calcute the median value every int(2*sigma+1)
    axis : int, optional
        The axis of `input` along which to smooth. Default is -1.
    '''
    if 0 in arr.shape:
        warnings.warn("input array is empty")
        return arr
    if method == 'gaussian_fft':
        kernel1d = signal.windows.gaussian(2*(4*sigma)+1, sigma)
        return convolve1d(arr, kernel1d, axis=axis)
    elif method == 'boxcar':
        kernel1d = np.ones(int(2*sigma+1))
        return convolve1d(arr, kernel1d, axis=axis)
    elif method == 'gaussian':
        return ndimage.gaussian_filter1d(arr, sigma, axis=axis)
    elif method == 'median':
        return median_filter_1d(arr, size=int(2*sigma+1), axis=axis)
    else:
        raise(ValueError("method not supports"))

def smooth1d_fft(x, ys, Tcut, axis=-1):
    """
    x : array_like shape:(m,)
    ys : array_like
        The input array.
    Tcut : components with period smaller than Tcat will be removed.
    """
    if len(x) != ys.shape[axis]:
        raise ValueError(f"len(x) needs equal to and ys.shape[axis]")
    n = len(x)
    Fy = np.fft.rfft(ys, n=n, norm='ortho', axis=axis)
    f = np.fft.rfftfreq(len(x), x[1]-x[0])
    Fy = np.moveaxis(Fy, axis, -1)
    Fy[...,f > 1/Tcut] *= 0
    Fy = np.moveaxis(Fy, -1, axis)
    return np.fft.irfft(Fy, n=n, norm='ortho', axis=axis) # need specify n when ifft or irfft



###############
# Author
# Qingze Chen
###############

__all__ = ['parser', 'IO']

# Cell
from hifast.utils.io import *
import numpy as np
from spectral_cube import SpectralCube
from tqdm import tqdm
import numpy as np
#from scipy.ndimage import rotate
#nbdev_comment _all_ = ['parser']

# Internal Cell
sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class = formatter_class, allow_abbrev=False,
                        description='Stray-radiation correction', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
parser.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                    help="if set, don't check or add ra dec")

parser.add_argument('--fitspath', required=True,
                   help='the path of the FITS file.')
# parser.add_argument('--pat', '--pattern_fpath', 
#                    help='path of the file stored pattern of 19-beams.')
parser.add_argument('--ra_range', type=float,nargs=2, default=[0, float('inf')],
                   help='ra range; unit: deg')
parser.add_argument('--dec_range', type=float,nargs=2, default=[0, float('inf')],
                   help='dec_range; unit: deg')




###### load fits cube ############
class in_fits(object):
    def __init__(self,f_file):
        hdu=fits.open(f_file)
        self.hdu=fits.open(f_file)
        i=self.hdu[0].header
        self.ARefP=i['CRPIX1']
        self.DRefP=i['CRPIX2']
        self.VRefP=i['CRPIX3']
        self.UnitA=i['CDELT1']
        self.UnitD=i['CDELT2']
        self.UnitV=i['CDELT3']
        self.StartA=i['CRVAL1']
        self.StartD=i['CRVAL2']
        self.StartV=i['CRVAL3']
        self.numV=i['NAXIS3']
        self.FitsDat=hdu[0].data
        self.CUnit3=i['CUNIT3']
###### load pattern and ita ############
class pattern_ita(object):
    def __init__(self, p_file=os.path.dirname(__file__) + '/core/data/pattern_eta.npz'):
        PI=np.load(p_file)
        self.ita=PI['eta']
        self.pattern=PI['pattern']

# Cell
class IO(BaseIO):
    ver = 'old'
    def _get_fpart(self,):
        return '-srmod'
    
    def _get_rotate_angle(self):
        args = self.args
        # use M01 M08
        m01_path = args.fpath.replace(f'M{self.nB:02d}', 'M01')
        m08_path = args.fpath.replace(f'M{self.nB:02d}', 'M08')

        from .ripple.util import Read_hdf5, Args
        h1 = Read_hdf5(Args(m01_path))
        h8 = Read_hdf5(Args(m08_path))

        from astropy.coordinates import SkyCoord
        from astropy import units as u
        coord1 = SkyCoord(h1.ra[0], h1.dec[0], unit = (u.deg, u.deg))
        coord8 = SkyCoord(h8.ra[0], h8.dec[0], unit = (u.deg, u.deg))
        
        # position_angle is the angle with N longitude
        rot_angle = 90 - coord1.position_angle(coord8).deg
        
        self.angle = None
        for a in [23.4, 53.4, 0. ]:
            if np.abs(rot_angle - a) < 1:
                self.angle = a
        
        if self.angle is None:
            raise ValueError("Only support multibeam rotation angle 23.4, 53.4 and 0.")
        else:
            print("Rotation angle:", self.angle)
         
    def rotate(self,img):
        args = self.args
        
        self._get_rotate_angle()
        
        angle = (90+self.angle)*np.pi/180  #23.4&53.4//53.4 is the true direction. 180 deg is for substraction
        # 读取库图片 Attention 转换 默认为int8 运算时可能会溢出
        # 设置新的图像大小
        h,w = img.shape[0],img.shape[1]
        newW = int(w*abs(np.cos(angle)) + h*abs(np.sin(angle)))+1
        newH = int(w*abs(np.sin(angle)) + h*abs(np.cos(angle)))+1
        # Attention dtype
        newimg1 = np.zeros((newH,newW,3))
        newimg2 =  np.zeros((newH,newW,3))
        newimg3 =  np.zeros((newH,newW,3))
        # 设置旋转矩阵 scr -> dex
        trans1 = np.array([[1,0,0],[0,-1,0],[-0.5*w,0.5*h,1]])
        trans1 = trans1.dot(np.array([[np.cos(angle),-np.sin(angle),0],[np.sin(angle),np.cos(angle),0],[0,0,1]]))
        trans1 = trans1.dot(np.array([[1,0,0],[0,-1,0],[0.5*newW,0.5*newH,1]]))
        # des -> src
        trans2 = np.array([[1,0,0],[0,-1,0],[-0.5*newW,0.5*newH,1]])
        trans2 = trans2.dot(np.array([[np.cos(angle),np.sin(angle),0],[-np.sin(angle),np.cos(angle),0],[0,0,1]]))
        trans2 = trans2.dot(np.array([[1,0,0],[0,-1,0],[0.5*w,0.5*h,1]]))
        # 开始旋转
        for x in range(w):
            for y in range(h):
                newPos = np.array([x,y,1]).dot(trans1)
                newimg1[int(newPos[1])][int(newPos[0])] = img[y][x]

        for x in range(newW):
            for y in range(newH):
                srcPos = np.array([x,y,1]).dot(trans2)
                if srcPos[0] >= 0 and srcPos[0] < w and srcPos[1] >= 0 and srcPos[1] < h:
                    # 双线性内插
                    bix,biy = int(srcPos[0]),int(srcPos[1]) # 取左上角坐标
                    # 避免最后一行溢出
                    if bix < w-1 and biy < h-1:
                        # 沿 X 方向线性内插
                        rgbX1 = img[biy][bix] + (img[biy][bix+1] - img[biy][bix])*(srcPos[0]-bix)
                        rgbX2 = img[biy+1][bix] + (img[biy+1][bix+1] - img[biy+1][bix])*(srcPos[0]-bix)
                        # 沿 Y  方向内插
                        rgb = rgbX1 + (rgbX2-rgbX1)*(srcPos[1]-biy)
                        newimg3[y][x] = rgb
        return newimg3[:,:,2]
    
    def func(self, x, y):
        if x<=39 and x>=0 and y<=39 and y>=0:
            return self.kernel[int(x)][int(y)]
        else:
            return 0
        
    def rotatefunc(self, x, y):
        if x<=39 and x>=0 and y<=39 and y>=0:
            return self.rotate(self.kernel)[int(x)][int(y)]
        else:
            return 0
    #SPfunc=np.vectorize(rotatefunc)
    def func2(self, x, y):
        #b=get_nB(self.args.fpath)-1
        r1=self.rotate(self.kernel)
        r2=r1[7:47,7:47]
        if x<=39 and x>=0 and y<=39 and y>=0:
            return r2[int(x)][int(y)]
        else:
            return 0
    #SPfunc=np.vectorize(rotatefunc)
    @property
    def SPfunc(self,):
        return np.vectorize(self.func2)

################################
    def vel_loop(self,n_channel,de,al,OpCube,dec,Ta,p,mpix,sk,C_start,ita,r2):
        krange=13
        if mpix==10:
            krange=20
        for c in range(n_channel-1):
            # print('       ',c)
            cor=0
            for k in range(krange):#Dec DIRECTION
                if de[k]<0 or de[k]>39:
                    cor=cor
                else:
                    for l in range(int(krange)):#R.A DIRECTION
                        #print(al[l])
                        if al[l]<=0 or al[l]>39:
                            cor=cor
                        else:
                        #print(SPfunc(de[l],al[k]))   test
                            cor=cor+OpCube[c][k][l]*r2[al[l],de[k]]
                #multiple with 3 because of the pixel size
                    #SSprint(cor)
                    #multiple with 3 because of the pixel size
                    #cor=cor+OpCube[c][k][l]*newfunc(20-3*np.cos(dec[b,1270+p]*np.pi/180)*(13-k)-BiasA, 3*l-1-BiasD) #old version
            if cor<0 or Ta[p,C_start+c]<0:
                #pointscube[0][p][c]=Ta[p,C_start+c]+0.0
                self.pointscube[0][p][c]=(Ta[p,C_start+c]/ita-cor/ita*((1-ita)*mpix*mpix*np.cos(dec[p]*np.pi/180)/sk))
            else:  
                self.pointscube[0][p][c]=(Ta[p,C_start+c]/ita-cor/ita*((1-ita)*mpix*mpix*np.cos(dec[p]*np.pi/180)/sk))#normalize it 


    def cut_spec_num(self):
        args = self.args
        s2p = self.s2p[:]
        Ta0=self.s2p
        # merge polar
        if len(Ta0.shape) == 3:
            Ta0 = np.nanmean(Ta0, axis = -1)
            
        ra_range = args.ra_range
        dec_range = args.dec_range
        freq_range = args.frange

        ra0=self.ra
        dec0=self.dec
        vel=self.vel
        freq=self.freq

        okno = (ra0 > ra_range[0]) & (ra0 < ra_range[1]) & (dec0 > dec_range[0]) & (dec0 < dec_range[1])
        self.ra=ra0[okno]
        self.dec=dec0[okno]
        self.s2p=Ta0[okno, :]
        self.mjd = self.mjd[okno]
        
        return okno
                
#############################
    def gen_s2p_out(self,):
        args = self.args
        
        # cut spec in ra dec range
        okno = self.cut_spec_num()
        
        # load fits
#############################################################        
        fitspath=args.fitspath
        fits_cube=in_fits(fitspath)
        self.PI=pattern_ita()
        ARefP=fits_cube.ARefP
        DRefP=fits_cube.DRefP
        VRefP=fits_cube.VRefP
        UnitA=fits_cube.UnitA
        UnitD=fits_cube.UnitD
        UnitV=fits_cube.UnitV
        StartA=fits_cube.StartA
        StartD=fits_cube.StartD
        StartV=fits_cube.StartV
        numV=fits_cube.numV
        FitsDat=fits_cube.FitsDat
        hdu=fits.open(fitspath)
        SCcube = SpectralCube.read(fitspath).with_spectral_unit(u.km / u.s) 
        fitswcs=WCS(hdu[0].header)
#to function 

########################################
        
        
        #### parameters #########               
        P_start=0# the start of points
        n_point=len(self.ra) # the number of points
        
        #  spectral axis cut
        vel = self.vel
        fitsvel=SCcube.spectral_axis.value
        
        ch0=min(fitsvel[0],vel[0])
        ch1=max(fitsvel[-1],vel[-1])
        self.is_use_freq = is_use_vel = (vel >= ch1) & (vel <= ch0)
        print("Spectral range:", ch1, ch0)
        self.is_use_freq2 = is_use_vel
        velv = vel[is_use_vel]
        
        C_start=(np.arange(len(vel))[is_use_vel])[0]         # the start of channels
        n_channel=len(velv) 
        
########### interrupt when channel of velocity not consistent
        if np.abs(fitsvel[1]-fitsvel[0]-(velv[1]-velv[0]))>0.01:
            print("The channel of velocity not consistent")
            print("fits:"+str(fitsvel[1]-fitsvel[0]))
            print("hdf5:"+str((vel[1]-vel[0])))

            raise(ValueError("The channel of velocity not consistent"))
            
        #########################
        self.pointscube=np.zeros((1,n_point,n_channel))# beams,points,channels
        beam_start = time.perf_counter()
        
        b = self.nB - 1
        ita=self.PI.ita[b]
        self.kernel=kernel=self.PI.pattern[b]
        r1=self.rotate(kernel)
        sumk=np.sum(kernel)
        r2=r1[7:47,7:47]
        x = np.linspace(0,38,39) 
        y = np.linspace(0,38,39) 
        x, y = np.meshgrid(x, y)
     #   arcmin30=IO.SPfunc(x,y)          
    #    newfunc = interpolate.interp2d(x, y, arcmin30, kind='cubic')#newfunc为一个函数 


        if np.abs(UnitD)>0.024 and np.abs(UnitD)<0.026:
            print("pix size is 90 arcsec")
            mDel=7
            mAlf=7
            mpix=3
        elif np.abs(UnitD)>0.016 and np.abs(UnitD)<0.017:
            print("pix size is 60 arcsec") 
            mDel=10
            mAlf=10
            mpix=2
        else:
            raise(ValueError('size of pixels not support'))
        
        ra = self.ra
        dec = self.dec
        Ta = self.s2p
        
        iter_ = tqdm(range(n_point), desc='CPU 0: ', mininterval=3)
        for p in iter_:#p:point
    #        print('    ',p)
            NAlf=(ra[P_start+p]-StartA)/UnitA   #how many pixels from this point to ref point
            NDel=(dec[P_start+p]-StartD)/UnitD
            Alf=int(np.round(fitswcs.world_to_pixel_values(ra[P_start+p],dec[P_start+p], 0)[0]))      
            Del=int(np.round(fitswcs.world_to_pixel_values(ra[P_start+p],dec[P_start+p], 0)[1]))
            BiasA=NAlf+ARefP-Alf
            BiasD=NDel+DRefP-Del
            #print('fits start at',int(round((velv[0]-fitsvel[0])/UnitV)))
            #print('fits ends at',int(round((velv[-1]-fitsvel[-1])/UnitV)))
            Refcube=FitsDat[int(round((velv[0]-fitsvel[0])/UnitV)):int(round((velv[0]-fitsvel[0])/UnitV))+n_channel,Del-mDel:Del+mDel,Alf-mAlf:Alf+mAlf]
            #Refcube=FitsDat[int(round((velv[0]-fitsvel[0])/UnitV)):int(round((velv[0]-fitsvel[0])/UnitV))+n_channel,Del-mDel:Del+mDel,Alf-mAlf:Alf+mAlf] #Cube:[freq,dec,ra]
            OpCube0=copy.deepcopy(Refcube)
            OpCube=np.nan_to_num(OpCube0)
            #%%
            #print(np.shape(OpCube))
            #%%
            if(Del<mDel or Del>np.shape(FitsDat)[1]-mDel or Alf<mAlf or Alf>np.shape(FitsDat)[2]-mAlf):
                OpCube=np.zeros([n_channel,mDel*2,mAlf*2])   
            al=[int(20-mpix*(l-mAlf)) for l in range(mAlf*2+1)]
            de=[int(20-mpix*(k-mDel)) for k in range(mDel*2+1)] 

#            if p%200==0:
#                print ("point="+str(p)+"   time=t+"+str((np.round(time.perf_counter()-beam_start,4))))
            self.vel_loop(n_channel,de,al,OpCube,dec,Ta,p,mpix,sumk,C_start,ita,r2)
            
        self.s2p_out = self.pointscube.transpose(1, 2, 0)
    
    
    def __call__(self, save=True):
        self.gen_s2p_out()
        self.gen_dict_out()
        # save to hdf5 file
        if save:
            self.save()
          #

# Cell
if __name__ == '__main__':
    import time
    import numpy as np
    from glob import glob
    from scipy.stats import binned_statistic_2d
    from astropy.io import fits
    from astropy import wcs
    import h5py
    import matplotlib.pyplot as plt
    from scipy import interpolate
    from PIL import Image
    from scipy import ndimage
    import copy
    from mpl_toolkits.mplot3d import Axes3D 
    from astropy.wcs import WCS
    from astropy.coordinates import SkyCoord  # High-level coordinates
    from astropy.coordinates import ICRS, Galactic, FK4, FK5  # Low-level frames
    from astropy.coordinates import Angle, Latitude, Longitude  # Angles
    import astropy.units as u
    ##########################################
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print(parser.format_values())  # useful for logging where different settings came from
    print(args_)
    io = IO(args_)
    io()




# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/ripple/markRFI.ipynb (unless otherwise specified).

__all__ = ['rms', 'get_rms_frange', 'real_rms', 'std', 'real_std', 'find_local_peak', 'fit_line', 'get_startend',
           'find_center', 'check_repeat_center', 'polyfit1order', 'mask_RFI', 'find_edge_2sides', 'mask_RFI_2sides',
           'center_theory', 'find_RFI', ]

# Cell
# author: Xu Chen 2021.06

import sys
import os

from astropy import log
from matplotlib import pyplot as plt

import numpy as np
import h5py
from copy import deepcopy
from tqdm import tqdm

# Cell
def rms(data,vel,rms_vrange=None):
    """
    data: 1D or 2D numpy array
    vel: velocity or frequency
    rms_vrange: [a,b], calculate rms from a to b
    """
    if rms_vrange is not None:
        rms_vrange.sort()
        is_use = (vel > rms_vrange[0])&(vel < rms_vrange[1])
        if len(data.shape) == 1:
            data = data[is_use]
        elif len(data.shape) ==2:
            data = data[:,is_use]

    ans = np.sqrt(np.nanmean((data)**2,axis = -1))
    return ans

def get_rms_frange(spec,freq,rms_step=None,is_excluded = None):
    if is_excluded is None: is_excluded = np.full(len(freq),False)
    RMSl = []
    rms_vrangel = []
    for v in np.arange(freq[0],freq[-1],rms_step):
        rms_vrange=[v,v+rms_step]
        use = (freq > rms_vrange[0]) & (freq < rms_vrange[1])
        if not np.any(use & is_excluded):
            RMS_ = rms(spec,freq,rms_vrange)
            if RMS_>0:
                RMSl += [RMS_,]
                rms_vrangel += [rms_vrange,]
    RMSl = np.array(RMSl); rms_vrangel = np.array(rms_vrangel)

    if len(RMSl) < 1:
        raise ValueError("Too much RFI, try a smaller rms_step or set a certain range by human!")

    RMS = np.nanmedian(RMSl)
    loc = np.argmin(np.abs(RMSl-RMS))
    rms_range = rms_vrangel[loc]
    log.info(f"Redirected to rms_range = {rms_range}.")

    return rms_range

def real_rms(data,vel,sigma,rms_vrange=None):
    """
    data: 1D or 2D numpy array
    vel: velocity or frequency
    rms_vrange: [a,b], calculate rms from a to b
    sigma: gaussian_filter1d sigma
    """
    from scipy.ndimage import gaussian_filter1d
    if rms_vrange is not None:
        is_use = (vel > rms_vrange[0])&(vel < rms_vrange[1])
        if len(data.shape) == 1:
            data = data[is_use]
        elif len(data.shape) ==2:
            data = data[:,is_use]

    g = gaussian_filter1d(data, sigma, axis = -1)
    ans = np.sqrt(np.nanmean((data - g)**2,axis = -1))

    return ans

def std(data,vel,vrange=None):
    """
    data: 1D or 2D numpy array
    vel: velocity or frequency
    vrange: [a,b], calculate std from a to b
    """
    if vrange is not None:
        is_use = (vel > vrange[0])&(vel < vrange[1])
        if len(data.shape) == 1:
            data = data[is_use]
        elif len(data.shape) ==2:
            data = data[:,is_use]

    ans = np.std(data,axis = -1)
    return ans

def real_std(data,vel,sigma,vrange=None):
    """
    data: 1D or 2D numpy array
    vel: velocity or frequency
    vrange: [a,b], calculate std from a to b
    sigma: gaussian_filter1d sigma
    """
    from scipy.ndimage import gaussian_filter1d
    if vrange is not None:
        is_use = (vel > vrange[0])&(vel < vrange[1])
        if len(data.shape) == 1:
            data = data[is_use]
        elif len(data.shape) ==2:
            data = data[:,is_use]

    g = gaussian_filter1d(data, sigma, axis = -1)
    ans = np.std(data - g,axis = -1)

    return ans

# Cell
def find_local_peak(rfi1,freq1,RMS,distance=5,):
    """
    rfi1,freq1: part of a spec and freq
    distance: signal.find_peaks distance

    return max freq and peak value
    """
    from scipy import signal
    loc_max = signal.find_peaks(rfi1,distance = distance)[0]

    loc_shape = loc_max.shape[0]
    if loc_shape == 1:
        loc = loc_max
    elif loc_shape > 1:
        loc = loc_max[np.argmax(rfi1[loc_max])]
    elif loc_shape == 0:
#         print("No local peak is found")
        return np.nan,np.nan

    freq_max = np.float(freq1[loc])
    flux_peak = np.float(rfi1[loc] - RMS)

    return freq_max,flux_peak

# Cell
def fit_line(rfi,freq,half_factor,**kwargs):
    """
    deleted
    """

    freq_peak,flux_peak = find_local_peak(rfi,freq,**kwargs)
    if np.sum(np.isnan([freq_peak,flux_peak])) > 0:
        return np.nan,np.nan

    limd,limu = flux_peak*.15,flux_peak*.85
    half_peak = flux_peak * half_factor

    left_use = (freq > freq_peak)&(rfi > limd)&(rfi < limu)
    righ_use = (freq < freq_peak)&(rfi > limd)&(rfi < limu)

    freq_left = freq[left_use]
    flux_left = rfi[left_use]
    freq_righ = freq[righ_use]
    flux_righ = rfi[righ_use]

    try:
        funcl = np.polyfit(flux_left,freq_left,1)
        f_left = np.poly1d(funcl)
        funcr = np.polyfit(flux_righ,freq_righ,1)
        f_righ = np.poly1d(funcr)
    except TypeError as TE:
        #print(TE)
        return np.nan,np.nan

    half_v_left = f_left(half_peak)
    half_v_righ = f_righ(half_peak)

    if (half_v_left <= min(freq))|(half_v_righ >= max(freq)):
        return np.nan,np.nan

    return half_v_left, half_v_righ

# Cell
def get_startend(is_rfi,rfi_width_lim = None, ext_sec= 0, exclude = True):
    """
    get the start and end index of the continued True value in a bool array
    is_rfi: 1D bool array
    rfi_width_lim: int, width should above an int channel number
    ext_sec: int, extend edges channel number
    """
    if not np.any(is_rfi): return [], []

    N = len(is_rfi)
    if not exclude: N -= 1

    starend = np.diff(is_rfi.astype('int8'), prepend=np.int8(0), append=np.int8(0))
    start_ = np.where(starend==1)[0]
    end_ = np.where(starend==-1)[0]

    if end_[0] < start_[0]:
        start_ = np.insert(start_,0,0)
    if end_[-1] < start_[-1]:
        end_ = np.insert(end_,-1,N)

    if rfi_width_lim is None:
        start = start_ - ext_sec
        end = end_ + ext_sec
    else:
        if np.isscalar(rfi_width_lim):
            starend_use = (end_-start_)>rfi_width_lim
        elif len(rfi_width_lim) == 2:
            starend_use = ((end_-start_)>rfi_width_lim[0])&((end_-start_)<rfi_width_lim[1])
        else:
            raise(ValueError('rfi_width_lim should be a scalar or list with two elements'))
        if not starend_use.any():
            #log.warning("No True meets width condition.")
            return [],[]

        start = start_[starend_use] - ext_sec
        end = end_[starend_use] + ext_sec

    if min(start) < 0:  start[start<0] = 0
    if max(end) > N: end[end>N] = N

    return start,end

# Cell

def find_center(spec,freq,is_rfi,RMS,freq_thr = .5,freq_step = 8.1,check = True,
                rfi_groups = 'two_groups',**kwargs):
    """
    find period RFI
    spec,freq,is_rfi: 1D array
    freq_thr: estimate error when looking for peaks to polyfit
    freq_step: rfi period MHz
    check: check repeat center?
    rfi_groups: divide rfi into 2 or 3 groups
    **kwargs: in get_startend
    """

    start,end = get_startend(is_rfi,**kwargs)
    # big rfi position
    fc0 = [];peaks = []
    for s,e in zip(start,end):
        rfi1 = spec[s:e]
        freq1 = freq[s:e]
        freq_peak,flux_peak = find_local_peak(rfi1,freq1,RMS=RMS)
        if not np.isnan(freq_peak):
            vc = freq_peak
            fc0 += [vc,]; peaks += [flux_peak,]

    peaks = np.array(peaks)
    # from the biggest one to guess rfi set 1
    fc00 = fc0[np.argmax(peaks)]

    def rfi_set1(fc0,fc00,freq,freq_step,freq_thr,peaks):
        fcenter1 = []; peak1 = []
        former = int((fc00 - freq[0])//freq_step + 10)
        fc1 = np.arange(fc00-freq_step*former,1440,freq_step)

        fcenter2_loc = []
        ii = -1
        for j in range(len(fc0)):
            vc = fc0[j]
            ii += 1
            if np.min(np.abs(vc - fc1)) < freq_thr:
                # record rfi set 1 center
                fcenter1 += [vc,]; peak1 += [peaks[j],]
            else:
                fcenter2_loc += [int(ii),]

        fcenter1 = np.array(fcenter1)
        if check:
            peak1 = np.array(peak1)
            _,fcenter1 = check_repeat_center(fcenter1,freq_step=8.1,x_start = '1st point',peaks = peak1)

        return fcenter1,fcenter2_loc,fc1

    fcenter1,fcenter2_loc,fc1 = rfi_set1(fc0,fc00,freq,freq_step,freq_thr,peaks)

    if len(fcenter1)<2:
        log.warning(f"rfi set 1 only has {len(fcenter1)} items, please check if it is a high-flux source or non-period RFI. Try again...")
        peaks[np.argmax(peaks)] = 0
        # find another one
        fc00 = fc0[np.argmax(peaks)]
        fcenter1,fcenter2_loc,fc1 = rfi_set1(fc0,fc00,freq,freq_step,freq_thr,peaks)

    # guess rfi set 2 besides set 1
    peak_use = np.full(len(peaks),False)
    peak_use[fcenter2_loc] = True
    peaks[~peak_use] = 0
    fc01 = fc0[np.argmax(peaks)]

    def rfi_set2(fc0,fc1,fc01,freq,freq_step,freq_thr,peaks):
        fcenter2 = []; peak2 = []
        former = int((fc01 - freq[0])//freq_step + 10)
        fc2 = np.arange(fc01-freq_step*former,1440,freq_step)

        fcenter3_loc = []
        ii = -1
        for j in range(len(fc0)):
            vc = fc0[j]
            ii += 1
            if np.min(np.abs(vc - fc2)) < freq_thr:
                fcenter2 += [vc,]; peak2 += [peaks[j],]
            elif (np.min(np.abs(vc - fc1)) >= freq_thr)&(np.min(np.abs(vc - fc2)) >= freq_thr):
                fcenter3_loc += [int(ii),]

        fcenter2 = np.array(fcenter2)
        if check:
            peak2 = np.array(peak2)
            _,fcenter2 = check_repeat_center(fcenter2,freq_step=8.1,x_start = '1st point',peaks = peak2)

        return fcenter2,fcenter3_loc

    fcenter2,fcenter3_loc = rfi_set2(fc0,fc1,fc01,freq,freq_step,freq_thr,peaks)

    if len(fcenter2)<2:
        log.warning(f"rfi set 2 only has {len(fcenter2)} items, please check if it is a high-flux source or non-period RFI. Try again...")
        peaks[np.argmax(peaks)] = 0
        fc01 = fc0[np.argmax(peaks)]
        # find another one
        fcenter2,fcenter3_loc = rfi_set2(fc0,fc1,fc01,freq,freq_step,freq_thr,peaks)

    fcenter3 = []
    if rfi_groups == 'three_groups':
        # guess rfi set 3 besides set 1 & 2
        peak_use = np.full(len(peaks),False)
        peak_use[fcenter3_loc] = True
        peaks[~peak_use] = 0

        if (peaks == 0.).all() == False:
            fc02 = fc0[np.argmax(peaks)]

            former = int((fc02 - freq[0])//freq_step + 10)
            fc3 = np.arange(fc02-freq_step*former,1440,freq_step)
            peak3 = []
            for j in range(len(fc0)):
                vc = fc0[j]
                if np.min(np.abs(vc - fc3)) < freq_thr:
                    fcenter3 += [vc,]; peak3 +=[peaks[j]]
            fcenter3 = np.array(fcenter3)
            if check:
                if fcenter3.size > 0:
                    peak3 = np.array(peak3)
                    _,fcenter3 = check_repeat_center(fcenter3,freq_step=8.1,x_start = '1st point',peaks = peak3)

    fcenter3 = np.array(fcenter3)
    fc0 = np.array(fc0).flatten()

    return fcenter1,fcenter2,fcenter3,fc0

# Cell
def check_repeat_center(item,freq_step=8.1,x_start = '1st point',peaks = None):
    """
    item: 1D array
    freq_step: rfi period MHz
    x_start: 'zero' or '1st point'
    peaks: help to delete smaller peaks
    """
    if len(item) <= 0:
        raise ValueError(f"Input item length is {len(item)}!")
    if x_start == 'zero':
        x = np.around(item/freq_step)
    elif x_start == '1st point':
        x = np.around((item-item[0])/freq_step)

    if len(set(x)) < len(x):
        from collections import Counter
        c = Counter(x)
        repeat_loc = np.array(list(c.values())) > 1
        repeat_key = np.array(list(c.keys()))[repeat_loc]

        if isinstance(peaks,np.ndarray):
            delete_loc = []
            for i in range(len(repeat_key)):
                repeat_loc_in_x = np.where(x == repeat_key[i])[0]
                peak = np.full(len(x),-1)
                peak[repeat_loc_in_x] = peaks[repeat_loc_in_x]
                Max = np.max(peak)
                rep = np.sum(peak == Max)
                if rep > 1:
                    repeat_peak = np.where(peak == np.max(peak))[0]
                    del_loc = np.hstack((repeat_peak[:-1],
                                        np.where((peak >= 0)&(peak < np.max(peak)))[0]))
                else:
                    del_loc = np.where((peak >= 0)&(peak < Max))[0]
                delete_loc.append(del_loc)
            delete_loc = np.hstack(delete_loc)
        else:
            repeat_loc_in_x = np.array([np.where(x == repeat_key[i])[0] for i in range(len(repeat_key))])
            delete_loc = repeat_loc_in_x[:,1:].flatten()

        x = np.delete(x,delete_loc,)
        item = np.delete(item,delete_loc,)
        #log.warning("RFIs are too closed. So x has repeated item, delete {len(delete_loc)}.")
        #raise ValueError(f" RFIs are too closed. So x has repeated item, delete {len(delete_loc)}.")
    return x,item


def polyfit1order(item,plot = False,pdf = None,**kwargs):
    """
    linear polyfit
    """
    x,item = check_repeat_center(item,**kwargs)

    if len(x) <= 1:
        raise ValueError(f" x only has {len(x)} item, can't polyfit.")

    pfit = np.polyfit(x,item,1)
    pfunc = np.poly1d(pfit)
    if plot:

        if pdf is not None:
            plt.switch_backend('agg')
        fig,ax = plt.subplots()
        ax.plot(x,pfunc(x),label = f'k = {pfunc[1]:.3f}')
        ax.plot(x,item,'x',label = f'b = {pfunc[0]:.2f}')
        ax.legend()
        ax.grid()
        if pdf is not None:
            pdf.savefig();plt.close()
    return pfunc

# Cell

def mask_RFI(freq,is_rfi,theory,mask_all_theory = False,freq_from_theory = None,**kwargs):
    """
    theory: freqs should have period rfis
    mask_all_theory: mask all theory freqs?
    freq_from_theory: mask width from theory center
    """
    # fix width
    rfi_width_lim = kwargs['rfi_width_lim']
    ext_sec = kwargs['ext_sec']

    start,end = get_startend(is_rfi,rfi_width_lim=rfi_width_lim/2,ext_sec= ext_sec, exclude=False)
    bigrfi = np.full(is_rfi.shape[0],False)
    for s,e in zip(start,end):
        fs,fe = freq[s],freq[e]
        mark = (theory > fs) & (theory < fe)
        if mark.any() == True:
            bigrfi[s:e] = True

    if mask_all_theory:
        bigrfi2 = np.zeros_like(bigrfi,dtype = 'bool')
        for th in theory:
            bigrfi2[(freq>th - freq_from_theory/2) & (freq<th + freq_from_theory/2)] = True
        bigrfi3 = bigrfi | bigrfi2
    else:
        bigrfi3 = bigrfi

    return bigrfi,bigrfi3

def find_edge_2sides(spec,vel,peak_position,step,rms_thresh,Print=False,small_rfi_times=2,ext_times = 0):
    """
    find edge from center to two sides
    small_rfi_times: small rfi below, eg.2*RMS will not be masked
    """
    start = deepcopy(peak_position)
    part_rms = rms(spec,vel,rms_vrange=[start-step/2,start+step/2])
    if Print: print("peak in ",start,"initial rms:", part_rms, "rms_thresh:", rms_thresh)

    if part_rms < rms_thresh * small_rfi_times:
        if Print: print(f"initial rms {part_rms} < rms_thresh {rms_thresh} * {small_rfi_times}")
        edge = np.array([np.nan,np.nan])
        return edge

    while part_rms > rms_thresh:
        start += -step
        if start < min(vel):
            break
        part_rms = rms(spec,vel,rms_vrange=[start-step/2,start+step/2])
        if Print: print("left step rms:", part_rms)

    end = deepcopy(peak_position)
    part_rms = rms(spec,vel,rms_vrange=[end-step/2,start+end/2])
    while part_rms > rms_thresh:
        end += step
        if end > max(vel):
            break
        part_rms = rms(spec,vel,rms_vrange=[end-step/2,end+step/2])
        if Print: print("right step rms:", part_rms)

    edge = np.array([start,end]) + np.array([-1,1]) * ext_times * step
    edge[edge < min(vel)] = min(vel)
    edge[edge > max(vel)] = max(vel)
    if Print:
        print("start, end ",[start,end])
        print("final edge:", edge)
    return edge

def mask_RFI_2sides(spec,freq,is_rfi,theory0,mark_rfi_width,small_rfi_times,RMS,
                    chan_step=3,**kwargs):
    """
    chan_step:channel step when walk from center to two sides
    Other parameters are the same as previous.
    """

    flim_l = theory0 - mark_rfi_width/2
    flim_r = theory0 + mark_rfi_width/2
    fdelta = freq[1] - freq[0]
    '''
    rfi_width_lim = kwargs['rfi_width_lim']
    ext_sec = kwargs['ext_sec']

    start,end = get_startend(is_rfi,rfi_width_lim=rfi_width_lim/2,ext_sec= ext_sec,exclude=False)
    bigrfi = np.full(is_rfi.shape[0],False)
    for s,e in zip(start,end):
        fs,fe = freq[s],freq[e]
        mark = (theory0 > fs) & (theory0 < fe)
        if mark.any() == True:
            bigrfi[s:e] = True
    '''
    bigrfi2 = np.full(is_rfi.shape[0],False)
    for j in range(len(flim_l)):
        rfi_part = (freq > flim_l[j])&(freq < flim_r[j])
        rspec = deepcopy(spec)[rfi_part]

        edge = find_edge_2sides(rspec,freq[rfi_part],peak_position = theory0[j],small_rfi_times = small_rfi_times,
                         step=chan_step*fdelta,rms_thresh=RMS,Print=False)
        if edge[0] < edge[1]:
            edge_use = (freq >= edge[0])&(freq <= edge[1])
            bigrfi2[edge_use] = True
    bigrfi = bigrfi2

    return bigrfi,bigrfi2

def center_theory(freq,fcenter1,fcenter2,fcenter3,freq_step,plot,pdf,rfi_groups = 'two_groups'):
    """
    fcenter1,fcenter2,fcenter3: three groups centers
    Other parameters are the same as previous.
    """
    if fcenter3.size <= 1:
        rfi_groups = 'two_groups'

    if (rfi_groups == 'two_groups')|(rfi_groups == 'three_groups'):
        pfunc1 = polyfit1order(fcenter1, freq_step=freq_step, plot = plot, pdf = pdf)
        pfunc2 = polyfit1order(fcenter2, freq_step=freq_step, plot = plot, pdf = pdf)

        former = int((pfunc1[0] - freq[0])//pfunc1[1] + 10)
        theory1 = np.arange(pfunc1[0]-pfunc1[1]*former,1460,pfunc1[1])
        theory1 = np.delete(theory1,np.where((theory1 < freq[0])|(theory1 > freq[-1]))[0])
        former = int((pfunc2[0] - freq[0])//pfunc2[1] + 10)
        theory2 = np.arange(pfunc2[0]-pfunc2[1]*former,1460,pfunc2[1])
        theory2 = np.delete(theory2,np.where((theory2 < freq[0])|(theory2 > freq[-1]))[0])

        if rfi_groups == 'three_groups':
            try:
                pfunc3 = polyfit1order(fcenter3, freq_step=freq_step, plot = plot, pdf = pdf)
                former = int((pfunc3[0] - freq[0])//pfunc3[1] + 10)
                theory3 = np.arange(pfunc3[0]-pfunc3[1]*former,1460,pfunc3[1])
                theory3 = np.delete(theory3,np.where((theory3 < freq[0])|(theory3 > freq[-1]))[0])
            except ValueError:
                theory3 = np.array([])
            theory0 = np.hstack((theory1,theory2,theory3))
        elif rfi_groups == 'two_groups':
            theory3 = np.array([])
            theory0 = np.hstack((theory1,theory2))
        theory0.sort()

    elif rfi_groups == 'all':
        former = int((pfunc0[0] - freq[0])//pfunc0[1] + 10)
        pfunc0 = polyfit1order(fc0, freq_step=freq_step/2)
        theory0 = np.arange(pfunc0[0]-pfunc0[1]*4*former,1460,pfunc0[1])
        theory0 = np.delete(theory0,np.where((theory0 < freq[0])|(theory0 > freq[-1]))[0])
        theory1,theory2,theory3 = [],[],[]

    return theory0,theory1,theory2,theory3



def find_RFI(spec,freq,is_rfi,freq_step=8.1,RMS = None,freq_thr = 0.5, ext_edge = 0,
             plot = False,pdf = None,ylim = None,rfi_groups = 'two_groups',
             mask_RFI_method = 'fixed_freq',small_rfi_times = 2,chan_step = 3,
             mask_all_theory = False,freq_from_theory = None,mask_thr = 15,
             ret_type = 'all theory',**kwargs):
    """
    ext_edge : channel number
    mask_RFI_method: 'fixed_freq','2_sides'
    mask_thr: above ~ times of rms threshold will be masked
    Other parameters are the same as previous.
    ret_type: 'all theory','theory 123'
    """
    if plot:

        if pdf is not None:
            plt.switch_backend('agg')

    fdelta = freq[1] - freq[0]

    fcenter1,fcenter2,fcenter3,fc0 = find_center(spec,freq,is_rfi,RMS,freq_step=freq_step,
                                                 freq_thr =freq_thr,rfi_groups = rfi_groups,**kwargs)
    theory0,theory1,theory2,theory3 = center_theory(freq,fcenter1,fcenter2,fcenter3,freq_step =freq_step,
                            plot = plot,pdf = pdf,rfi_groups =rfi_groups)

    is_rfi_ = (spec > RMS * mask_thr)
    if mask_RFI_method == '2_sides':
        bigrfi,bigrfi2 = mask_RFI_2sides(spec,freq,is_rfi,theory0,mark_rfi_width = freq_from_theory,
                                 small_rfi_times = small_rfi_times,chan_step = chan_step,RMS = RMS,**kwargs)
    elif mask_RFI_method == 'fixed_freq':
        bigrfi,bigrfi2 = mask_RFI(freq,is_rfi_,theory0,mask_all_theory,freq_from_theory,**kwargs)

    if ext_edge > 0:
        from ..utils.misc import extend_Trues
        bigrfi2 = extend_Trues(bigrfi2,axis = -1,ext_add = ext_edge)

    if plot:
        fig,ax = plt.subplots(figsize=(15,3))
        ax.plot(freq,spec)
        spec_ = np.full(spec.shape[0],np.nan)
        spec_[bigrfi] = spec[bigrfi]
        ax.plot(freq,spec_)

        ymin = np.min(spec);ymax = np.max(spec)*.8

        ax.vlines(fc0,ymin=ymin,ymax=ymax,linestyles='--',colors='k',label = 'all')
        ax.vlines(fcenter1,ymin=ymin,ymax=ymax,linestyles='--',colors='r',label = '1')
        ax.vlines(fcenter2,ymin=ymin,ymax=ymax,linestyles='--',colors='g',label = '2')
        if (rfi_groups == 'three_groups') & (fcenter3.size>0):
            ax.vlines(fcenter3,ymin=ymin,ymax=ymax,linestyles='--',colors='b',label = '3')
        if ylim is not None:
            ax.set_ylim(ylim[0],ylim[1])
        ax.set_title('RFI positions coarse measure')
        plt.grid();plt.legend()
        if pdf is not None:
            pdf.savefig();plt.close()

        fig,ax = plt.subplots(figsize=(15,3))
        ax.plot(freq,spec)
        spec_ = np.full(spec.shape[0],np.nan)
        spec_[bigrfi2] = spec[bigrfi2]
        ax.plot(freq,spec_)
        if (rfi_groups == 'two_groups')|(rfi_groups == 'three_groups'):
            ax.vlines(theory1,ymin=ymin,ymax=ymax,linestyles='--',colors='r',label = 'theory1')
            ax.vlines(theory2,ymin=ymin,ymax=ymax,linestyles='--',colors='g',label = 'theory2')
            if (rfi_groups == 'three_groups') & (theory3.size>0):
                ax.vlines(theory3,ymin=ymin,ymax=ymax,linestyles='--',colors='b',label = 'theory3')
        elif rfi_groups == 'all':
            ax.vlines(theory0,ymin=ymin,ymax=ymax,linestyles='--',colors='k',label = 'theory0')
        if ylim is not None:
            ax.set_ylim(ylim[0],ylim[1])
        ax.set_title('RFI positions prediction')
        plt.grid();plt.legend()
        if pdf is not None:
            pdf.savefig();plt.close()

    if ret_type == 'all theory':
        return bigrfi2,theory0
    elif ret_type == 'theory 123':
        return bigrfi2,(theory1,theory2,theory3)





__all__ = ['read_regions', 'box', 'line', 'replace_region', 'mask_srcs']


import warnings
import math
import numpy as np


def read_regions(fpath):
    with open(fpath, 'r') as f:
        # skip comment
        line = f.readline().strip()
        while line:
            if line.startswith('#'):
                line = f.readline().strip()
            else:
                break

        if not line.startswith('global'):
            warnings.warn(f"format error in {fpath}, the first line not started with '#' should start with 'global'. Only DS9 format is supported, please check {fpath}")
            return None
        # region unit
        line = f.readline().strip()
        if not line:
            warnings.warn(f"no region found in {fpath}")
            return None
        elif line.strip() not in ['physical', 'image']:
            warnings.warn(f"only 'pixel' type regions are supported, no regions are read from {fpath}, please check {fpath}")
            return None
        # read region
        regions = {}
        line = f.readline().strip()
        while line:
            key, remain = line.split('(', 1)
            reg = tuple(map(float, remain.split(')', 1)[0].split(',')))
            if key in regions:
                regions[key] += [reg, ]
            else:
                regions[key] = [reg, ]
            line = f.readline().strip()
        return regions


def box(center_x, center_y, size_x, size_y, PA=None):
    """
    box: (center_x, center_y, size_x, size_y, PA?)
    """
    bottom_left = center_x - size_x/2, center_y - size_y/2
    top_right = center_x + size_x/2, center_y + size_y/2

    bottom_left = [max(math.ceil(num), 0) for num in bottom_left]
    top_right = [max(math.floor(num), 0) for num in top_right]

    return slice(bottom_left[0], top_right[0]+1), slice(bottom_left[1], top_right[1]+1)


def line(start_x, start_y, end_x, end_y):
    """
    line(start_x, start_y, end_x, end_y)
    """

    dx_abs = abs(start_x - end_x)
    dy_abs = abs(start_y - end_y)

    if dx_abs > dy_abs:
        s = max(math.ceil(min(start_x, end_x)), 0)
        e = max(math.floor(max(start_x, end_x)), 0)
        return slice(s, e+1), slice(None)
    else:
        s = max(math.ceil(min(start_y, end_y)), 0)
        e = max(math.floor(max(start_y, end_y)), 0)
        return slice(None), slice(s, e+1)


def replace_region(arr, regions, fill_value=np.nan, inplace=True):
    """
    arr: array; ndim >=2, slice the first two dim to replace
    regions:
    fill_value:
    inplace:
    """

    if not inplace:
        arr = np.copy(arr)
    for key in regions.keys():
        if key == 'box':
            for reg in regions[key]:
                arr[box(*reg)[::-1]] = fill_value # x axis is freq
        elif key == 'line':
            for reg in regions[key]:
                arr[line(*reg)[::-1]] = fill_value
        else:
            warnings.warn(f'region type {key} not support')
    return arr


from . import corr_vel as cv
from astropy.coordinates import SkyCoord


def mask_srcs(fpath, is_exclued, ra, dec, freq, mjd=None, inplace=True, rest_frame='LSRK'):
    """
    fpath: file inculde src position. example:
            #ra[deg], dec[deg], R[arcmin], freq_min[MHz], freq_max[MHz]
            349.63, 30.96, 20, 1300, 1302
            349.6510, 30.96226, 10, 1310, 1311
    is_exclued:
    ra, dec: in deg
    mjd:
    inplace: True change is_exclude inplace.
    rest_frame: if not 'none', fix rest frame of freq
    """
    if not inplace:
        is_exclued = np.copy(is_exclued)
    srcs = np.loadtxt(fpath, comments='#', delimiter=',', ndmin=2)
    cata = SkyCoord(ra=ra, dec=dec, unit='deg')
    for src in srcs:
        c = SkyCoord(src[0], src[1], unit='deg')
        ind = np.where(cata.separation(c).arcmin < src[2])[0]
        if len(ind) > 0:
            if rest_frame != 'none':
                freq_C = cv.frame_correct_freq(freq, ra[ind[0]], dec[ind[0]], mjd[ind[0]], frame=rest_frame)
            else:
                freq_C = freq
            ind_freq = np.where((freq_C >= src[3]) & (freq_C <= src[4]))[0]
            # freq is in ascending
            if len(ind_freq) > 0:
                is_exclued[ind, ind_freq[0]:ind_freq[-1]+1] = True
    return is_exclued





__all__ = ['IO']


from .utils.io import *
import copy


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                        description='downsample.', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input spectra file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
group = parser.add_argument_group(f'*downsample\n{sep_line}')
group.add_argument('--chan_factor', type=int, default = 0,
                    help='down sample on channel')
group.add_argument('--spec_factor', '--t_factor',type=int, default=0,
                    help='down sample on spec')


class IO(BaseIO):
    ver = 'new'

    def __init__(self, args, dict_in=None, inplace_args=False):
        """
        args: class
              including attributes: fpath, outdir, frange
        dict_in: if None, load data from args.fpath, if set, omit data in args.fpath
        """
        args = args if inplace_args else copy.deepcopy(args)
        args.no_radec = True
        super().__init__(args, dict_in=dict_in, inplace_args=inplace_args)

    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-ds'
        return fpart

    def down(self, arr, n, axis=-1, drop=True):
        from .utils.misc import average_every_n
        if n <= 1:
            return arr
        else:
            return average_every_n(arr, n, axis=axis, drop=drop).astype(arr.dtype)

    def __call__(self, save=True):

        from .core.radec import _tight_ra
        import scipy.interpolate as interp

        args = self.args
        type1 = ['is_delay', 'is_extrapo', 'is_on', 'mjd', 'next_to_cal']
        type2 = ['freq', 'vel']
        ## to do: if vel is under optical

        dict_out = {}
        dict_out["Header"] = self.Header
        for key in self.fs.keys():
            if key in type1:
                dict_out[key] = self.down(self.fs[key][()], args.spec_factor, axis=0)
            elif key in type2:
                if self.is_use_freq is not None:
                    dict_out[key] = self.fs[key][()][self.is_use_freq]
                else:
                    dict_out[key] = self.fs[key][()]
                dict_out[key] = self.down(dict_out[key], args.chan_factor, axis=0)
            elif key == 'Tcal':
                if self.is_use_freq is not None:
                    dict_out[key] = self.fs[key][()][:, self.is_use_freq]
                else:
                    dict_out[key] = self.fs[key][()]
                dict_out[key] = self.down(dict_out[key], args.chan_factor, axis=1)
            elif key == self.infield:
                dict_out[key] = self.down(self.down(self.s2p, args.spec_factor, axis=1), args.chan_factor, axis=2)
            elif key == 'is_rfi':
                value = self.fs[key][()]
                if self.is_use_freq is not None:
                    value = value[:, self.is_use_freq]
                dict_out[key] = self.down(self.down(value, args.spec_factor, axis=0), args.chan_factor, axis=1)
        if 'ra' in self.fs.keys():
            if args.spec_factor > 1:
                ra = _tight_ra(self.fs['ra'][()])
                dict_out['ra'] = interp.interp1d(self.fs['mjd'], ra, kind='linear', fill_value ='extrapolate')(dict_out['mjd'])
                dict_out['ra'][dict_out['ra']<0] += 360
            else:
                dict_out['ra'] = self.fs['ra'][()]
        if 'dec' in self.fs.keys():
            if args.spec_factor > 1:
                dict_out['dec'] = interp.interp1d(self.fs['mjd'], self.fs['dec'], kind='linear', fill_value ='extrapolate')(dict_out['mjd'])
            else:
                dict_out['dec'] = self.fs['dec'][()]

        self.dict_out = dict_out
        if save:
            self.save()


if __name__ == '__main__':
    args_ = parser.parse_args()
    print('#'*35+'Args'+'#'*35)
    print(parser.format_values())
    print('#'*35+'####'+'#'*35)
    io = IO(args_)
    io()





__all__ = ['sep_line', 'parser']


from .utils.io import *
from glob import glob
import json


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=formatter_class, allow_abbrev=False,
                        description='cut FAST RAW Data', )

parser.add_argument('fpath',
                    help='only need input the first chunk file path (e.g. XXX_0001.fits) of RAW spectra data')
parser.add_argument('-f', dest='force', action='store_true',
                    help='overwriting file if out file exists')
parser.add_argument('--outdir', required=True,
                   help='the directory to store output files.')
parser.add_argument('-g', is_write_out_config_file_arg=True,
                        help='save config to file path')
parser.add_argument('-c', '--my-config', is_config_file_arg=True,
                        help='config file path')
parser.add_argument('--frange', type=float, nargs=2,
                    help='freq range')
parser.add_argument('--step', type=int, default=1,
                   help='number of files processed every time, default 1')
parser.add_argument('--start', type=int, default=1,
                   help='chunk number of start')
parser.add_argument('--stop', type=int,
                   help='chunk number of stop')
parser.add_argument('--sep_save', type=bool_fun, choices=[True, False], default='False',
                   help='every step save to a file')
parser.add_argument('--h5_compression', default='none',
                   help='`compression` in `h5py create_dataset`; `none`, `lzf`, `gzip`, or a number in range(10). '
                         'Note: compressed file may not be opened by carta.')


if __name__ == '__main__':
    args_ = parser.parse_args()
    print('#'*35+'Args'+'#'*35)
    print(parser.format_values())
    print('#'*35+'####'+'#'*35)

    args = args_

    # check h5_compression
    try:
        args.h5_compression = int(args.h5_compression)
    except ValueError:
        pass

    #record history
    header = rec_his(args=json.dumps(args.__dict__))

    #check input file exits
    if not os.path.exists(args.fpath):
        raise(OSError(f'File {args.fpath} not exists.'))

    # replace patten in outdir
    nB = get_nB(args.fpath)
    project = get_project(args.fpath)
    ## use the dirname of the fits file as "date"
    date = os.path.basename(os.path.dirname(os.path.abspath(args.fpath)))
    args.outdir = sub_patten(args.outdir, date=date, nB=f'{nB:02d}', project=project)
    print(f'outdir: {args.outdir}')
    if args.outdir is not None:
        if not os.path.exists(args.outdir):
            print(f'outdir {args.outdir} not exists. Create it now')
            os.makedirs(args.outdir, exist_ok=True)
    ## check out file
    fname_add = date
    fname_part = re.sub('[0-9]{4}\.fits\Z', '', args.fpath)
    out_name_base = os.path.join(args.outdir, f"{os.path.basename(fname_part)}")

    fileout = out_name_base + rf"0001.hdf5"
    fileout_sep = glob(out_name_base + rf"[0-9][0-9][0-9][0-9].hdf5")


    if os.path.exists(fileout) or len(fileout_sep)>0:
        if args.force:
            print(f"will overwrite the existing out file")
        else:
            print(f"File exists {fileout}")
            print(fileout_sep)
            print('exit... Using -f to overwrite it.')
            sys.exit()
    # run
    from .core.cal2 import FASTRawCut
    Fd = FASTRawCut(args.fpath,
                    start=args.start,
                    stop=args.stop,
                    frange=args.frange,
                    verbose=True)
    Fd(outdir=args.outdir,
       step=args.step,
       header=header,
       sep_save=args.sep_save,
       h5_compression=args.h5_compression,
      )





__all__ = ['IO', 'interp_fun', 'align_calc']


from .utils.io import *


sep_line = '##'+'#'*70+'##'
parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}", formatter_class=formatter_class, allow_abbrev=False,
                        description='', )
add_common_argument(parser)
parser.add_argument('fpath',
                    help='input spectra temperature file path.')
parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')],
                    help='Limit frequence range')
parser.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                    help="if set, don't check or add ra dec")
group = parser.add_argument_group(f'*Flux\n{sep_line}')
# --flux affect the outfield name
group.add_argument('--final_fpath',
                   help='')
group.add_argument('--flux_fpath',
                   help='')
group.add_argument('--flux_fpath_bef',
                   help='')
group.add_argument('--T_fpath',
                   help='')


class IO(BaseIO):
    ver = 'old'

    def _get_fpart(self,):
        """
        need modify this function
        """
        fpart = '-fW'
        return fpart

    def _import_m(self,):
        """
        need modify this function
        """
        super()._import_m()
        global HFSpec, np, interp1d
        from .funcs import HFSpec
        import numpy as np
        from scipy.interpolate import interp1d

    def _process_fpaths(self,):
        args = self.args
        project = get_project(args.fpath)
        date = get_date_from_path(args.fpath)
        nB = get_nB(args.fpath)

        for pname in ['flux_fpath',
                     'flux_fpath_bef',
                     'final_fpath',
                     'T_fpath']:
            path = getattr(args, pname)
            path = sub_patten(path, date=date, nB=f'{nB:02d}', project=project)
            path = os.path.expanduser(path)
            setattr(args, pname, path)

    def gen_s2p_out(self,):
        args = self.args

        self._process_fpaths()
        S_flux = HFSpec(args.flux_fpath)
        S_flux_bef = HFSpec(args.flux_fpath_bef)
        S_final = HFSpec(args.final_fpath)
        S_T = HFSpec(args.T_fpath)

        for S in [S_flux, S_flux_bef, S_final, S_T]:
            assert (S.mjd == self.mjd).all()
        # get gain
        G_freq, Jy_K = align_calc(S_flux.freq, S_flux.flux, S_flux_bef.freq, S_flux_bef.Ta, axis=1, fun=lambda a,b:a/b)
        # get baseline in flux
        freq_tmp, flux_tmp = align_calc(G_freq, Jy_K, S_T.freq, S_T.Ta, axis=1, fun=lambda a,b:a*b)
        bl_freq, bl_flux = align_calc(freq_tmp, flux_tmp, S_final.freq, S_final.flux, axis=1, fun=lambda a,b:a-b)
        del freq_tmp, flux_tmp

        # Temperature cali; only process 'merge' and 'div' for now
        print('Temperature Calibration')
        print(f'Using info from {args.T_fpath}')
        s2p_out = self.s2p / (interp_fun(S_T.freq, S_T.pcals_merged_s, axis=1)(self.freq) * S_T.pcals_amp_diff_interp_values[:][:, None, :])
        # power of W is 16 times of N ??
        s2p_out *= 16
        s2p_out[self.fs['is_on']] -= 1
        s2p_out *= interp_fun(S_T.freq, S_T.Tcal, axis=1)(self.freq)
        # flux cali
        print('Flux Calibration')
        print(f'Using info from {args.flux_fpath} \nand\n{args.flux_fpath_bef} ')
        s2p_out *= interp_fun(G_freq, Jy_K, axis=1)(self.freq)
        self.outfield = 'flux'
        del G_freq, Jy_K
        # Sub baseline
        print('Baseline and SW removal')
        s2p_out -= interp_fun(bl_freq, bl_flux, axis=1)(self.freq)
        self.s2p_out = s2p_out
        print('RFI')
        if 'is_rfi' in S_final.keys():
            is_rfi = interp1d(S_final.freq, S_final.is_rfi, axis=1, kind='nearest', bounds_error=False, fill_value=0)(self.freq)
            is_rfi = is_rfi.astype(bool)
        else:
            is_rfi = None
        self.is_rfi = is_rfi

    def __call__(self, save=True):
        args = self.args
        self.gen_s2p_out()
        self.gen_dict_out()

        if self.is_rfi is not None:
            self.dict_out['is_rfi'] = self.is_rfi
        if save:
            self.save()


def interp_fun(*args, **kwrags):
    from scipy.interpolate import interp1d
    return interp1d(*args, bounds_error=False, fill_value=np.nan, **kwrags)

def align_calc(freq1, arr1, freq2, arr2, axis=-1, fun=None):
    """
    freq1:
    """
    freq_range = [max(freq1.min(), freq2.min()), min(freq1.max(), freq2.max())]
    is_1 = (freq1 > freq_range[0]) & (freq1 < freq_range[1])
    is_2 = (freq2 > freq_range[0]) & (freq2 < freq_range[1])

    assert (freq1[is_1] == freq2[is_2]).all()
    freq_new = freq1[is_1]

    sli1 = [slice(None),] * arr1.ndim
    sli1[axis] = np.where(is_1)[0]
    sli2 = [slice(None),] * arr2.ndim
    sli2[axis] = np.where(is_2)[0]
    sli1 = tuple(sli1)
    sli2 = tuple(sli2)

    return freq_new, fun(arr1[sli1], arr2[sli2])


if __name__ == '__main__':
    args_ = parser.parse_args()
    # print(parser.format_help())
    # print("----------")
    print('#'*35+'Args'+'#'*35)
    args_from = parser.format_values()
    print(args_from)
    print('#'*35+'####'+'#'*35)

    HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
    io = IO(args_, HistoryAdd=HistoryAdd)
    io()



"""
hifast.radec
============

Calculate celestial coordinates (RA, DEC) for FAST drift scan data.

This module matches spectral data with feed cabin position data (KY files)
to compute the Right Ascension (RA) and Declination (DEC) for each spectrum.
It supports interpolation, coordinate transformation using `erfa` or `astropy`,
and environmental corrections.

It can also process a KY file directly to calculate and plot the feed cabin's
trajectory independent of spectral data.
"""

import os
import re
import sys
from glob import glob
import argparse
from .utils.io import *

def create_parser():
    sep_line = '##'+'#'*70+'##'
    parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                        formatter_class=argparse.ArgumentDefaultsHelpFormatter, allow_abbrev=False,
                        description='Calculate RA/DEC coordinates from feed cabin data.')
    
    # --- Input/Output ---
    group = parser.add_argument_group('Input/Output')
    group.add_argument('fname', metavar='FILE',
                        help='Input file path. Can be an HDF5 spectral data file (must contain "S/mjd") or a KY feed cabin log file (.xlsx).')
    group.add_argument('--outdir', metavar='DIR',
                        help='Output directory. Defaults to the input file directory (for HDF5) or current directory (for XLSX).')
    group.add_argument('-f', dest='force', action='store_true',
                        help='Force overwrite of the output file if it already exists.')
    group.add_argument('--plot', action='store_true',
                        help='Create a PDF plot showing the calculated RA/DEC trajectory.')

    # --- Configuration ---
    group = parser.add_argument_group('Configuration')
    group.add_argument('--ky_files', nargs='*', metavar='FILE',
                        help='Explicitly specify one or more KY (feed cabin log) file paths. Separate multiple files with spaces. If omitted, searches in `--ky_dir` based on observation time.')
    group.add_argument('--ky_dir', metavar='DIR',
                        help="Base directory to search for KY files. Defaults to standard locations, with `~/KY` checked first, followed by others like `/data31/KY`.")
    group.add_argument('--backend', choices=['erfa', 'astropy'], default='astropy',
                       help='Coordinate calculation backend. "astropy" (default) is recommended (auto-handles Earth orientation). "erfa" is a lower-level alternative.')
    group.add_argument('--tol', type=float, default=1, metavar='SEC',
                       help='Maximum allowed extrapolation time [seconds]. Coordinates are extrapolated for gaps smaller than this; larger gaps raise an error.')
    group.add_argument('--ky_fixed', action='store_true',
                       help='Assume fixed feed position. Useful for early drift scans with incomplete trajectory logs.')

    # --- Environment ---
    group = parser.add_argument_group('Environment')
    group.add_argument('--phpa', type=float, default=925., metavar='HPA',
                   help='Atmospheric pressure [hPa]. Used for refraction correction.')
    group.add_argument('--temperature', type=float, default=15., metavar='DEG_C',
                   help='Ground-level temperature [Celsius]. Used for refraction correction.')
    group.add_argument('--humidity', type=float, default=0.8, metavar='0-1',
                   help='Relative humidity [0.0 - 1.0]. Used for refraction correction.')
    group.add_argument('--dUT1', metavar='SEC',
                   help='UT1-UTC time difference [seconds]. Only used for "erfa" backend (ignored for "astropy" as it handles it automatically).')

    # --- Performance ---
    group = parser.add_argument_group('Performance')
    group.add_argument('-n', '--nproc', type=int, default=1, metavar='INT',
                       help='Number of parallel processes for multi-beam calculation.')
    group.add_argument('--no_cache', action='store_true',
                       help='Disable using cached intermediate results (forces fresh KY file processing).')
    
    return parser

parser = create_parser()

if __name__ == '__main__':
    args = parser.parse_args()
#     print('#'*35+'Args'+'#'*35)
#     print(parser.format_values())  # useful for logging where different settings came from
#     print('#'*35+'####'+'#'*35)

    fname = args.fname
    ky_files = args.ky_files
    tol = args.tol
    ky_fixed = args.ky_fixed
    use_cache = not args.no_cache
    plot = args.plot
    outdir = args.outdir
    nproc = args.nproc
    
    outpart = '-radec'
    if outdir is None:
        if fname[-5:] == '.xlsx':
            outdir = './'
        else:
            outdir = os.path.dirname(fname)
    fileout = os.path.join(outdir, '.'.join(os.path.basename(fname).split('.')[:-1]) + f'{outpart}.hdf5')
    if os.path.exists(fileout):
        if args.force:
            print(f"will overwrite the existing out file {fileout}")
        else:
            print(f"File exists {fileout}")
            print('exit... Using -f to overwrite it.')
            sys.exit()
            
    from .core.radec import plot_radec, get_radec
    if args.ky_dir is not None:
        from .core import radec as radec_py
        radec_py.ky_dir_default = [args.ky_dir]
    
    env_para = {}
    env_para['phpa'] = args.phpa
    env_para['temperature'] = args.temperature
    env_para['humidity'] = args.humidity
    try:
        dUT1 = float(args.dUT1)
    except:
        dUT1 = args.dUT1
    radec = get_radec(fname, ky_files=ky_files, tol=tol, ky_fixed=ky_fixed, use_cache=use_cache, nproc=nproc, 
                      backend=args.backend, env_para=env_para, dUT1=dUT1)
    if radec is None:
        raise(ValueError('The attempt to process RA-DEC was unsuccessful. Exiting the procedure...'))
        sys.exit(1)
    #saving
    ##record history
    from .utils.io import *
    import json
    header = rec_his(args=json.dumps(args.__dict__))
    print('Saving...')
    radec['Header'] = header
    save_specs_hdf5(fileout, radec)
    print(f"Saved to {fileout}")
    if plot:
        plot_radec(radec, fileout + '.pdf')





__all__ = ['parser', 'args', 'fname', 'obsfile', 'frange', 'noise_mode', 'noise_date', 'sigma', 'obsmode', 'modes',
           'CalinBs', 'outdir', 'calname', 'ccrd', 'fpparas', 't_src', 'ncir', 'saveT', 'processing_nB', 'outdata',
           'nBs_have']


from . import FLUXGAIN as cbr
from ..utils.io import save_dict_hdf5, replace_nB
from ..utils.io import bool_fun
from ..utils import obs_log
import numpy as np
import os
import sys
import h5py


parser = cbr.parser
parser.add_argument('--nproc', '--np', type=int, default=3,
                   help='number of process used')
parser.add_argument('--cache', type=bool_fun, choices=[True, False], default='False',
                       help='If set to True, will cache/load the results of all beams except for the initial one.')
parser.add_argument('--plot', action='store_true',
                       help='plot FluxGain with frequency.')
parser.add_argument('--obs_log',
                       help='If set, will use the obs_log to get `-d`, `-m`, `-n`, `--noise_mode`.\
                        It requires a file path ending in `.log` or `.txt`, or a directory path where the log file can be searched for.')


args = parser.parse_args()


if args.obs_log is not None:
    if not os.path.exists(args.obs_log):
        raise ValueError(f'`--obs_log` not exists: `{args.obs_log}`.')
    if os.path.isfile(args.obs_log):
        obs_log_path = args.obs_log
    elif os.path.isdir(args.obs_log):
        print(f'try to search for log file in the directory: `{args.obs_log}`.')
        obs_log_path = obs_log.search_log_fpath(args.fname, args.obs_log)
    else:
        raise ValueError(f'`--obs_log` should be a file path ending in `.log` or `.txt`, or a directory path where the log file can be searched for.\
        \nBut got `{args.obs_log}`.')
    print(f'Use obs_log file: `{obs_log_path}`.')
    paras = obs_log.get_sep_para_from_log(obs_log_path)[0]
    print(f'got paras: {paras}')
    args.d = paras['n_delay']
    args.m = paras['n_on']
    args.n = paras['n_off']
    args.noise_mode = paras['noise_mode']
elif args.d is None or args.m is None or args.n is None or args.noise_mode is None:
    raise ValueError(f'`--obs_log` is not set, but `--d`, `--m`, `--n`, `--noise_mode` are not set either.')


fname= args.fname
obsfile= fname.split('/')[-1]

d, m, n = args.d, args.m, args.n
frange= args.frange
noise_mode = args.noise_mode
noise_date = args.noise_date
sigma= args.smt_sigma
obsmode= args.obsmode

# may remove
modes= ['Drift','MultiBeamCalibration','OnOff','MultiBeamOTF','DriftWithAngle','DecDriftWithAngle']
if obsmode not in modes:
    print('Please check observation mode.')
    os._exit(0)
# end

CalinBs= args.nBs
if CalinBs== None:
    if obsmode in ['MultiBeamCalibration','MultiBeamOTF',]:
        nBs= np.arange(1,20)
    else:
        nBs= [1,]
else:
    nBs= CalinBs
print('Flux Calibration of ' + replace_nB(obsfile, 0), end='')
print(' of Beams ', nBs)


outdir= args.outdir
if outdir==None:
    outdir= '.'
calname= args.calname
ccrd = cbr.load_crd(calname,args.crd)
print('Calibrater position:', ccrd)
fpparas= args.fluxProfilePara

# for tracking
t_src= args.t_src/3600./24#change into days
ncir= args.n_cir
#T_select=args.T_select
#rlim_src = args.rlim_src
saveT= args.saveT



# set cbr.* global variable. Need improve later
cbr.fname = fname
cbr.d = d
cbr.m = m
cbr.n = n
cbr.frange = frange
cbr.noise_mode = noise_mode
cbr.noise_date = noise_date
#cbr.oname = oname
cbr.obsmode = obsmode
cbr.sigma = sigma
cbr.saveT = saveT
cbr.obsmode = obsmode
cbr.nBs = nBs
cbr.ccrd = ccrd
cbr.fpparas = fpparas
cbr.calname = calname
cbr.outdir = outdir
cbr.t_src = t_src
cbr.ncir = ncir
#cbr.T_select = T_select
#cbr.rlim_src = rlim_src
cbr.args = args


def processing_nB(nB, check_cache=False, outname_pattern=None):
    if check_cache:
        cache_name = outname_pattern + f'-M{nB:02d}-cache.h5'
        if os.path.exists(cache_name):
            print(f'Load FluxGain of beam {nB:02d} from cached file: {cache_name}')
            outdata = {k:v[()] for k,v in h5py.File(cache_name,'r').items()}
            for key in outdata.keys():
                if isinstance(outdata[key], bytes):
                    outdata[key] = outdata[key].decode('utf-8')
            return outdata
    outdata = {}
    i = nB
    cbr.nB = nB

    try:
        K_Jy,Tcal_s,tcal_file,mjd,ONp,OFFp,TMAX= cbr.load_data(fname)
    except SystemExit as e:
        print(f'Fail to process Beam {nB:02d}')
        return None

    print(fname,i)
    radec= cbr.radec
    ra,dec,ZD = radec[f'ra{i}'], radec[f'dec{i}'],radec[f'ZD{i}']
    outdata[f'M{nB:02d}']= K_Jy
    outdata[f'Tcal{nB}']= Tcal_s
    outdata[f'ra{i}']= ra
    outdata[f'dec{i}']= dec
    outdata[f'ZD{i}']= ZD
    outdata[f'ONp{i}']= ONp
    outdata[f'OFFp{i}']= OFFp
    outdata[f'Tmax{i}']= TMAX
    print(f'Finished calibration of M{nB:02d}, {cbr.fit_k*100:.3f}% input data has been ignored.')
    print('#'*50)

    # common
    outdata['Tcal_file']= tcal_file
    print('Used Tcal of '+tcal_file)
    outdata['freq']= cbr.freq
    outdata['cal_flux']= cbr.flux
    outdata['mjd']= mjd
    #
    outdata['outname'] = cbr.outname
    if check_cache:
        print(f'Save FluxGain of beam {nB:02d} to cache file {cache_name}')
        save_dict_hdf5(cache_name, outdata)
    return outdata


# process the first one, M01 in general, and the radec are generated
cbr.nproc = args.nproc
print(f'Process Beam {nBs[0]:02d}')
outdata = processing_nB(nBs[0])
if outdata is None:
    print('Fail to process the first beam in ``--nBs``. Abort.')
    sys.exit(1)
nBs_have = 1
if len(nBs) > 1:
    print(f'Process rest Beams with {args.nproc} CPUs')
    if args.cache:
        def processing_nB_with_cache(nB): return processing_nB(nB,True,cbr.outname)
    else:
        processing_nB_with_cache = processing_nB
    from multiprocessing import Pool
    with Pool(args.nproc) as pool:
        result = pool.map(processing_nB_with_cache, np.copy(nBs).tolist()[1:])

    for j, d in enumerate(result):
        if d is not None:
            outdata.update(d)
            nBs_have += 1
        else:
            print(f'Fail to process Beam {nBs[1:][j]:02d}')


if nBs_have == 19:
    print("19 beams have been successfully processed and saving to file with name containing 'All'.")
    outname = outdata['outname']+'-FluxGain-All.hdf5'
else:
    print("Not all 19 beams were processed successfully and saved to a file that doesn't have 'All' in its name.")
    outname= outdata['outname'] + '-FluxGain.hdf5'
save_dict_hdf5(outname,outdata)
print('Saved to '+ outname)
if args.plot:
    from .plot import plot
    fig = plot(outdata, suptitle=outname)
    fig.savefig(outname.split('.', 1)[0] + '.png')




"""
hifast.bld
==========

Baseline fitting and subtraction for spectral data.

This module provides tools to fit and subtract baselines from spectra. It supports various fitting methods (PLS, polynomial, spline) and preprocessing steps (smoothing, binning) along both time and frequency axes.
"""


from .utils.io import *


__all__ = ['IO', 'IO_i', 'create_parser', 'parser']

sep_line = '##'+'#'*70+'##'

def create_parser():
    parser = ArgumentParser(prog=f"python -m hifast.{os.path.basename(sys.argv[0])[:-3]}",
                            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                            allow_abbrev=False,
                            description='Fit and subtract the baseline from spectral data.')
    add_common_argument(parser)
    
    # Input/Output
    parser.add_argument('fpath', metavar='FILE',
                        help='Path to the input HDF5 file containing spectra (e.g., "data.hdf5").')
    parser.add_argument('--frange', type=float, nargs=2, default=[0, float('inf')], metavar=('START', 'END'),
                        help='Frequency range to process [MHz]. Default: process all.')
    parser.add_argument('--no_radec', action='store_true', not_in_write_out_config_file=True,
                        help="Skip verification/addition of RA/Dec coordinates.")
    parser.add_argument('--show_prog', type=bool_fun, choices=[True, False], default='True', env_var='HIFAST_SHOW_PROG',
                        help='Show progress bar during processing.')
    parser.add_argument('--nproc', '-n', type=int, default=1, metavar='N',
                        help='Number of parallel processes to use.')

    # Baseline Fitting Method
    group = parser.add_argument_group('Baseline Fitting Method')
    
    rew_type = ['asym1', 'asym2', 'asym3', 'sym1',]
    method = ['none']
    method += ['PLS-'+r for r in rew_type]
    method += ['poly-'+r for r in rew_type]
    method += ['knpoly-'+r for r in rew_type]
    method += ['Gauss-'+r for r in rew_type]
    method += ['knspline-'+r for r in rew_type]
    method += ['masPLS-'+r for r in rew_type]
    method += ['asPLS', 'arPLS']
    method += ['original']
    
    group.add_argument('--method', default='arPLS', choices=method,
                       help='Baseline fitting algorithm. "arPLS" is robust for most cases. "poly-*" uses polynomial fitting.')
    group.add_argument('--lam', type=float, default=1.0e8, metavar='LAMBDA',
                       help='Smoothing parameter (lambda) for PLS, Spline, and Gauss methods. Larger values = smoother (stiffer) baseline. Typical range: 1e4 - 1e9.')
    group.add_argument('--deg', type=int, default=2, metavar='DEG',
                       help='Degree for polynomial or spline methods.')
    group.add_argument('--knots', type=str, metavar='JSON_FILE',
                       help='Path to JSON file with knots for spline methods.')
    group.add_argument('--offset', type=float, default=2,
                       help='Offset parameter.')
    group.add_argument('--ratio', type=float, default=0.01,
                       help='Ratio parameter.')
    group.add_argument('--niter', type=int, default=100, metavar='N',
                       help='Maximum number of iterations.')
    group.add_argument('--exclude_add', '--exclude_type', default='none', choices=['none', 'auto1', 'auto2'],
                       help='Auto-exclusion method. "auto1": exclude low weights; "auto2": exclude outliers (>3 sigma).')

    # Preprocessing
    group = parser.add_argument_group('Preprocessing (Before Fitting)')
    group.add_argument('-T', '--trans', type=bool_fun, choices=[True, False], default='False',
                       help='Fit baseline along TIME axis instead of FREQUENCY axis. WARNING: This swaps the roles of time and frequency parameters.')
    group.add_argument('--njoin', '--njoin_t', '--average_every_t', type=int, default=0, metavar='N',
                       help='Average N adjacent time samples to increase SNR before fitting. Reduces time resolution of the baseline model.')
    group.add_argument('--s_method_t', default='none', choices=['none', 'gaussian', 'boxcar', 'median'],
                       help='Smoothing method along TIME axis to suppress noise. Options: gaussian, boxcar, median.')
    group.add_argument('--s_sigma_t', type=int, default=5, metavar='SIGMA',
                       help='Width parameter for time-axis smoothing (sigma for Gaussian, window size for others).')
    group.add_argument('--s_method_freq', default='none', choices=['none', 'gaussian', 'boxcar', 'median', 'PLS', 'fft'],
                       help='Smoothing method along FREQUENCY axis.')
    group.add_argument('--s_sigma_freq', type=float, default=5, metavar='SIGMA',
                       help='Width parameter for frequency-axis smoothing.')
    group.add_argument('--average_every_freq', '--njoin_freq', type=int, default=0, metavar='N',
                       help='Binning factor along FREQUENCY axis.')

    # Exclusion
    group = parser.add_argument_group('Exclusion Settings')
    group.add_argument('--use_pre_is_excluded', type=bool_fun, choices=[True, False], default='False',
                       help='Use existing "is_excluded" mask from input file (e.g., from previous RFI flagging).')
    group.add_argument('--src_file', metavar='CATALOG',
                       help='Path to source catalog file for masking known sources (columns: RA, Dec, Radius, MinFreq, MaxFreq).')
    group.add_argument('--frame', choices=['BARYCENT', 'HELIOCEN', 'LSRK', 'LSRD'], default='LSRK',
                       help='Velocity frame for source masking.')

    # Post-processing
    group = parser.add_argument_group('Post-processing (Optional)')
    
    rew_type = ['asym1', 'asym2', 'asym3', 'sym1',]
    method = ['none']
    method += ['poly-'+r for r in rew_type]
    
    group.add_argument('--post_method', default='none', choices=method,
                       help='Method for secondary baseline fitting on individual spectra (e.g., to remove residuals after time-averaged fitting).')
    group.add_argument('--post_s_method_freq', default='none', choices=['none', 'gaussian', 'boxcar', 'median', 'PLS', 'fft'],
                       help='Smoothing method for post-processing.')
    group.add_argument('--post_s_sigma_freq', type=float, default=5, metavar='SIGMA',
                       help='Smoothing sigma for post-processing.')
    group.add_argument('--post_average_every_freq', type=int, default=0, metavar='N',
                       help='Binning factor for post-processing.')
    group.add_argument('--post_deg', type=int, default=2, metavar='DEG',
                       help='Polynomial degree for post-processing.')
    group.add_argument('--post_offset', type=float, default=2,
                       help='Offset for post-processing.')
    group.add_argument('--post_ratio', type=float, default=0.01,
                       help='Ratio for post-processing.')
    group.add_argument('--post_niter', type=int, default=100, metavar='N',
                       help='Iterations for post-processing.')
    group.add_argument('--post_exclude_add', '--post_exclude_type', default='none', choices=['none', 'auto1', 'auto2'],
                       help='Exclusion method for post-processing.')

    # Interaction
    group = parser.add_argument_group('Interaction')
    group.add_argument('-i', '--interact', action='store_true', not_in_write_out_config_file=True,
                       help='Enable interactive mode.')
    group.add_argument('--ylim', nargs='+', default=['auto'], not_in_write_out_config_file=True,
                       help='Y-axis limits for plots (e.g., "auto" or "-1 5").')
    group.add_argument('--figsize', type=float, nargs=2, default=(10, 7), not_in_write_out_config_file=True, metavar=('W', 'H'),
                       help='Figure size in inches.')
    group.add_argument('--length', type=int, default=20, not_in_write_out_config_file=True, metavar='N',
                       help='Number of spectra to test in interactive mode.')
    group.add_argument('--start_init', type=int, default=0, not_in_write_out_config_file=True, metavar='INDEX',
                       help='Starting spectrum index.')
    
    return parser

parser = create_parser()


class IO(BaseIO):
    ver = 'old'

    def _get_fpart(self,):
        fpart = '-bld'
        if self.args.post_method is not None and self.args.post_method != 'none':
            fpart += '_p'
        return fpart

    def _import_m(self,):
        """
        Lazy import of heavy dependencies (numpy, h5py, etc.).
        """
        super()._import_m()
        global h5py, OrderedDict, sub_baseline, np

        import h5py
        import numpy as np

        from collections import OrderedDict
        from .core.baseline import sub_baseline

    @staticmethod
    def fit_baseline(s2p, freq, mjd, args, is_excluded=None):
        fit_kwargs = {}
        keys = ['nproc',
                'method',
                'njoin',
                's_method_t',
                's_sigma_t',
                's_method_freq',
                's_sigma_freq',
                'average_every_freq',
                'lam',
                'deg',
                'offset',
                'ratio',
                'niter',
                'exclude_add',
               ]
        for key in keys:
            fit_kwargs[key] = getattr(args, key)
        fit_kwargs['is_excluded'] = is_excluded

        if fit_kwargs['method'].startswith('knspline') or fit_kwargs['method'].startswith('knpoly'):
            # load knots from knots file
            import json
            with open(args.knots, 'r') as f:
                fit_kwargs['knots'] = json.load(f)

        trans = getattr(args, 'trans', False)
        if trans:
            if fit_kwargs['is_excluded'] is not None:
                 fit_kwargs['is_excluded'] = fit_kwargs['is_excluded'].transpose((1, 0, 2))
            return sub_baseline(np.arange(len(mjd)), s2p.transpose((1, 0, 2)), subtract=True, **fit_kwargs).transpose((1, 0, 2))
        else:
            return sub_baseline(freq, s2p, subtract=True, **fit_kwargs)

    @staticmethod
    def post_fit_baseline(s2p, freq, mjd, args, is_excluded=None):
        fit_kwargs = {}
        keys = ['method',
                's_method_freq',
                's_sigma_freq',
                'average_every_freq',
                'deg',
                'offset',
                'ratio',
                'niter',
                'exclude_add',
               ]
        for key in keys:
            fit_kwargs[key] = getattr(args, 'post_'+key)
        fit_kwargs['nproc'] = args.nproc
        fit_kwargs['njoin'] = 0
        fit_kwargs['s_method_t'] = 'none'
        fit_kwargs['s_sigma_t'] = 0
        fit_kwargs['lam'] = 0
        fit_kwargs['is_excluded'] = is_excluded

        trans = getattr(args, 'trans', False)
        if trans:
            if fit_kwargs['is_excluded'] is not None:
                 fit_kwargs['is_excluded'] = fit_kwargs['is_excluded'].transpose((1, 0, 2))
            return sub_baseline(np.arange(len(mjd)), s2p.transpose((1, 0, 2)), subtract=True, **fit_kwargs).transpose((1, 0, 2))

        return sub_baseline(freq, s2p, subtract=True, **fit_kwargs)

    def _load_is_rfi(self,):
        fs = self.fs
        if 'is_rfi' in fs.keys():
            is_rfi = fs['is_rfi'][:]
            if self.is_use_freq is not None:
                is_rfi = is_rfi[:, self.is_use_freq]

                # M06 bandpass is not flat in [1410, 1416]
            if self.nB == 6:
                whole_rfi = np.all(is_rfi, axis=1)
                freq = self.freq
                is_use = (freq > 1406) & (freq < 1416)
                is_rfi[:, is_use] = False
                is_rfi[whole_rfi] = True

            self.is_rfi = is_rfi

    def _load_is_excluded(self):
        if not self.args.use_pre_is_excluded:
            return
        fs = self.fs
        if 'is_excluded' in fs.keys():
            is_excluded = fs['is_excluded'][:]
            if self.is_use_freq is not None:
                is_excluded = is_excluded[:, self.is_use_freq]
            # M06 bandpass is not flat in [1410, 1416]
            if self.nB == 6:
                freq = self.freq
                is_use = (freq > 1406) & (freq < 1416)
                is_excluded[:, is_use] = False
            self.is_excluded = is_excluded

    def _load_sources(self):
        args = self.args
        fpath = args.src_file
        if fpath is not None:
            from .core.regions import mask_srcs
            is_excluded = self.is_excluded if hasattr(self,'is_excluded') else np.zeros(self.s2p.shape[:2], dtype = bool)
            print(f"loading excluded files from {fpath}")
            self.is_excluded = mask_srcs(fpath, is_excluded, self.ra, self.dec, self.freq,
                                         self.mjd, inplace=True, rest_frame=args.frame)

    def gen_s2p_out(self,):
        args = self.args

        # gen self.s2p_out
        s2p = self.s2p[:]
        # rfi not in is_excluded
        self._load_is_rfi()
        # is_excluded
        self._load_is_excluded()
        # src add to is_excluded
        self._load_sources()

        # fit baseline:
        if args.method is not None and args.method != 'none':
            if hasattr(self,'is_excluded'):
                is_excluded = self.is_excluded
                # suppress nan warning in baseline fitting in the future
                # whole_rfi = np.all(is_excluded, axis=1)
                # is_excluded[whole_rfi] = False # I don't like some warnings when whole spec is nan ...
                # check is_excluded shape
                if is_excluded.ndim == 2:
                    is_excluded = np.full(is_excluded.shape + s2p.shape[2:], is_excluded[..., None])
                else:
                    raise(ValueError('Now `is_excluded` only supports 2-dim.'))
            else:
                is_excluded = None

            print('fit and substract baseline')
            s2p = self.fit_baseline(s2p, self.freq, self.mjd, args, is_excluded)
            if args.post_method is not None and args.post_method !='none':
                print('applying low-order polynomial fitting on each spectrum')
                s2p = self.post_fit_baseline(s2p, self.freq, self.mjd, args, is_excluded)
        else:
            print('no baseline method assigned, skip. Make sure the input spectra have been baselined.')
        self.s2p_out = s2p

    def __call__(self, save=True):
        self.gen_s2p_out()
        self.gen_dict_out()

        if hasattr(self, 'is_excluded'):
            # update
            self.dict_out['is_excluded'] = self.is_excluded
        if hasattr(self, 'is_rfi'):
            # update
            self.dict_out['is_rfi'] = self.is_rfi

        # save to hdf5 file
        if save:
            self.save()


def check_backend():
    import matplotlib as mpl
    if 'ipympl' not in mpl.get_backend():
        print('Please use interaction mode in Jupyert and run \'%matplotlib ipympl\' in the notebook cell first')
        sys.exit()

interact_save_lim = 20
def interact(args):
    check_backend()
    import h5py
    from .interaction import bld_i as interact
    f = h5py.File(args.fpath, 'r')
    fs = f['S']
    if 'T' in fs.keys():
        T = fs['T']
    elif 'Ta' in fs.keys():
        T = fs['Ta']
    elif 'flux' in fs.keys():
        T = fs['flux']
    if args.use_pre_is_excluded and 'is_excluded' in fs.keys():
        is_excluded = fs['is_excluded']
    else:
        is_excluded = None
    interact.T2p = T
    interact.is_excluded = is_excluded
    interact.freq = fs['freq'][:]
    interact.frange = args.frange
    interact.nproc = args.nproc
    interact.length = min(args.length, interact.T2p.shape[1])
    interact.figsize = args.figsize
    interact.ylim = args.ylim[0] if len(args.ylim) == 1 else args.ylim
    interact.trans = args.trans
    interact.start_init = args.start_init
    if T.shape[1] <= interact_save_lim:
        interact.save = True
    else:
        interact.save = False
    return interact.main()
    # sys.exit()


class IO_i(IO):

    def set_interact_spec(self, interact_spec):
        self.interact_spec = interact_spec

    def gen_s2p_out(self,):
        args = self.args
        if args.interact:
            print('Please ensure you have adjusted the parameters for each `polar`, otherwise you will get nan vaules in the file')
            self.s2p_out = self.interact_spec.bld_out
            return


if __name__ == '__main__':
    args_ = parser.parse_args()
    if args_.interact:
        interact_spec, widgets = interact(args_)[0:2]
        save = IO_i(args_, HistoryAdd={'interact':str(widgets)})
        save.set_interact_spec(interact_spec)
        if interact_spec.save:
            print('Please run \'save()\' in the notebook cell to save your results')
    else:
        print('#'*35+'Args'+'#'*35)
        args_from = parser.format_values()
        print(args_from)
        print('#'*35+'####'+'#'*35)

        HistoryAdd = {'args_from': args_from} if args_.my_config is not None else None
        io = IO(args_, HistoryAdd=HistoryAdd)
        io()
